from __future__ import annotations

import numpy as np
import pytest

from glotaran.builtin.megacomplexes.damped_oscillation import DampedOscillationMegacomplex
from glotaran.builtin.megacomplexes.decay import DecayMegacomplex
from glotaran.builtin.megacomplexes.spectral import SpectralMegacomplex
from glotaran.model import Model
from glotaran.optimization.optimize import optimize
from glotaran.parameter import Parameters
from glotaran.project import Scheme
from glotaran.simulation import simulate

DampedOscillationsModel = Model.create_class_from_megacomplexes(
    [DampedOscillationMegacomplex, DecayMegacomplex, SpectralMegacomplex]
)


class OneOscillation:
    sim_model = DampedOscillationsModel(
        **{
            "megacomplex": {
                "m1": {
                    "type": "damped-oscillation",
                    "labels": ["osc1"],
                    "frequencies": ["osc.freq"],
                    "rates": ["osc.rate"],
                },
                "m2": {
                    "type": "spectral",
                    "shape": {"osc1_cos": "sh1", "osc1_sin": "sh1"},
                },
            },
            "shape": {
                "sh1": {
                    "type": "gaussian",
                    "amplitude": "shapes.amps.1",
                    "location": "shapes.locs.1",
                    "width": "shapes.width.1",
                },
            },
            "dataset": {"dataset1": {"megacomplex": ["m1"], "global_megacomplex": ["m2"]}},
        }
    )

    model = DampedOscillationsModel(
        **{
            "megacomplex": {
                "m1": {
                    "type": "damped-oscillation",
                    "labels": ["osc1"],
                    "frequencies": ["osc.freq"],
                    "rates": ["osc.rate"],
                },
            },
            "dataset": {"dataset1": {"megacomplex": ["m1"]}},
        }
    )

    wanted_parameter = Parameters.from_dict(
        {
            "osc": [
                ["freq", 25.5],
                ["rate", 0.1],
            ],
            "shapes": {"amps": [7], "locs": [5], "width": [4]},
        }
    )

    parameter = Parameters.from_dict(
        {
            "osc": [
                ["freq", 20],
                ["rate", 0.3],
            ],
        }
    )

    time = np.arange(0, 3, 0.01)
    spectral = np.arange(0, 10)
    axis = {"time": time, "spectral": spectral}

    wanted_clp = ["osc1_cos", "osc1_sin"]
    wanted_shape = (300, 2)


class OneOscillationWithIrf:
    sim_model = DampedOscillationsModel(
        **{
            "megacomplex": {
                "m1": {
                    "type": "damped-oscillation",
                    "labels": ["osc1"],
                    "frequencies": ["osc.freq"],
                    "rates": ["osc.rate"],
                },
                "m2": {
                    "type": "spectral",
                    "shape": {"osc1_cos": "sh1", "osc1_sin": "sh1"},
                },
            },
            "shape": {
                "sh1": {
                    "type": "gaussian",
                    "amplitude": "shapes.amps.1",
                    "location": "shapes.locs.1",
                    "width": "shapes.width.1",
                },
            },
            "irf": {
                "irf1": {
                    "type": "gaussian",
                    "center": "irf.center",
                    "width": "irf.width",
                },
            },
            "dataset": {
                "dataset1": {
                    "megacomplex": ["m1"],
                    "global_megacomplex": ["m2"],
                    "irf": "irf1",
                }
            },
        }
    )

    model = DampedOscillationsModel(
        **{
            "megacomplex": {
                "m1": {
                    "type": "damped-oscillation",
                    "labels": ["osc1"],
                    "frequencies": ["osc.freq"],
                    "rates": ["osc.rate"],
                },
            },
            "irf": {
                "irf1": {
                    "type": "gaussian",
                    "center": "irf.center",
                    "width": "irf.width",
                },
            },
            "dataset": {
                "dataset1": {
                    "megacomplex": ["m1"],
                    "irf": "irf1",
                }
            },
        }
    )

    wanted_parameter = Parameters.from_dict(
        {
            "osc": [
                ["freq", 25],
                ["rate", 0.1],
            ],
            "shapes": {"amps": [7], "locs": [5], "width": [4]},
            "irf": [["center", 0.3], ["width", 0.1]],
        }
    )

    parameter = Parameters.from_dict(
        {
            "osc": [
                ["freq", 25],
                ["rate", 0.1],
            ],
            "irf": [["center", 0.3], ["width", 0.1]],
        }
    )

    time = np.arange(0, 3, 0.01)
    spectral = np.arange(0, 10)
    axis = {"time": time, "spectral": spectral}

    wanted_clp = ["osc1_cos", "osc1_sin"]
    wanted_shape = (300, 2)


class OneOscillationWithSequentialModel:
    sim_model = DampedOscillationsModel(
        **{
            "initial_concentration": {
                "j1": {"compartments": ["s1", "s2"], "parameters": ["j.1", "j.0"]},
            },
            "k_matrix": {
                "k1": {
                    "matrix": {
                        ("s2", "s1"): "kinetic.1",
                        ("s2", "s2"): "kinetic.2",
                    }
                }
            },
            "megacomplex": {
                "m1": {"type": "decay", "k_matrix": ["k1"]},
                "m2": {
                    "type": "damped-oscillation",
                    "labels": ["osc1"],
                    "frequencies": ["osc.freq"],
                    "rates": ["osc.rate"],
                },
                "m3": {
                    "type": "spectral",
                    "shape": {
                        "osc1_cos": "sh1",
                        "osc1_sin": "sh1",
                        "s1": "sh2",
                        "s2": "sh3",
                    },
                },
            },
            "shape": {
                "sh1": {
                    "type": "gaussian",
                    "amplitude": "shapes.amps.1",
                    "location": "shapes.locs.1",
                    "width": "shapes.width.1",
                },
                "sh2": {
                    "type": "gaussian",
                    "amplitude": "shapes.amps.2",
                    "location": "shapes.locs.2",
                    "width": "shapes.width.2",
                },
                "sh3": {
                    "type": "gaussian",
                    "amplitude": "shapes.amps.3",
                    "location": "shapes.locs.3",
                    "width": "shapes.width.3",
                },
            },
            "irf": {
                "irf1": {
                    "type": "gaussian",
                    "center": "irf.center",
                    "width": "irf.width",
                },
            },
            "dataset": {
                "dataset1": {
                    "initial_concentration": "j1",
                    "irf": "irf1",
                    "megacomplex": ["m1", "m2"],
                    "global_megacomplex": ["m3"],
                }
            },
        }
    )

    model = DampedOscillationsModel(
        **{
            "initial_concentration": {
                "j1": {"compartments": ["s1", "s2"], "parameters": ["j.1", "j.0"]},
            },
            "k_matrix": {
                "k1": {
                    "matrix": {
                        ("s2", "s1"): "kinetic.1",
                        ("s2", "s2"): "kinetic.2",
                    }
                }
            },
            "megacomplex": {
                "m1": {"type": "decay", "k_matrix": ["k1"]},
                "m2": {
                    "type": "damped-oscillation",
                    "labels": ["osc1"],
                    "frequencies": ["osc.freq"],
                    "rates": ["osc.rate"],
                },
            },
            "irf": {
                "irf1": {
                    "type": "gaussian",
                    "center": "irf.center",
                    "width": "irf.width",
                },
            },
            "dataset": {
                "dataset1": {
                    "initial_concentration": "j1",
                    "irf": "irf1",
                    "megacomplex": ["m1", "m2"],
                }
            },
        }
    )

    wanted_parameter = Parameters.from_dict(
        {
            "j": [
                ["1", 1, {"vary": False, "non-negative": False}],
                ["0", 0, {"vary": False, "non-negative": False}],
            ],
            "kinetic": [
                ["1", 0.2],
                ["2", 0.01],
            ],
            "osc": [
                ["freq", 25],
                ["rate", 0.1],
            ],
            "shapes": {"amps": [0.07, 2, 4], "locs": [5, 2, 8], "width": [4, 2, 3]},
            "irf": [["center", 0.3], ["width", 0.1]],
        }
    )

    parameter = Parameters.from_dict(
        {
            "j": [
                ["1", 1, {"vary": False, "non-negative": False}],
                ["0", 0, {"vary": False, "non-negative": False}],
            ],
            "kinetic": [
                ["1", 0.2],
                ["2", 0.01],
            ],
            "osc": [
                ["freq", 25],
                ["rate", 0.1],
            ],
            "irf": [["center", 0.3], ["width", 0.1]],
        }
    )

    time = np.arange(-1, 5, 0.01)
    spectral = np.arange(0, 10)
    axis = {"time": time, "spectral": spectral}

    wanted_clp = ["osc1_cos", "osc1_sin", "s1", "s2"]
    wanted_shape = (600, 4)


@pytest.mark.parametrize(
    "suite",
    [
        OneOscillation,
        OneOscillationWithIrf,
        OneOscillationWithSequentialModel,
    ],
)
def test_doas_model(suite):
    print(suite.sim_model.validate())
    assert suite.sim_model.valid()

    print(suite.model.validate())
    assert suite.model.valid()

    print(suite.sim_model.validate(suite.wanted_parameter))
    assert suite.sim_model.valid(suite.wanted_parameter)

    print(suite.model.validate(suite.parameter))
    assert suite.model.valid(suite.parameter)

    dataset = simulate(suite.sim_model, "dataset1", suite.wanted_parameter, suite.axis)
    print(dataset)

    assert dataset.data.shape == (suite.axis["time"].size, suite.axis["spectral"].size)

    print(suite.parameter)
    print(suite.wanted_parameter)

    data = {"dataset1": dataset}
    scheme = Scheme(
        model=suite.model,
        parameters=suite.parameter,
        data=data,
        maximum_number_function_evaluations=20,
    )
    result = optimize(scheme, raise_exception=True)
    print(result.optimized_parameters)

    for param in result.optimized_parameters.all():
        assert np.allclose(param.value, suite.wanted_parameter.get(param.label).value, rtol=1e-1)

    resultdata = result.data["dataset1"]
    assert np.array_equal(dataset["time"], resultdata["time"])
    assert np.array_equal(dataset["spectral"], resultdata["spectral"])
    assert dataset.data.shape == resultdata.fitted_data.shape
    assert np.allclose(dataset.data, resultdata.fitted_data)

    assert "damped_oscillation_cos" in resultdata
    assert "damped_oscillation_sin" in resultdata
    assert "damped_oscillation_associated_spectra" in resultdata
    assert "damped_oscillation_phase" in resultdata



"""Glotaran module with utilities for ``ipython`` integration (e.g. ``notebooks``)."""
from __future__ import annotations

from collections import UserString

from glotaran.plugin_system.io_plugin_utils import infer_file_format
from glotaran.typing.types import StrOrPath


class MarkdownStr(UserString):
    """String wrapper class for rich display integration of markdown in ipython."""

    def __init__(self, wrapped_str: str, *, syntax: str | None = None):
        """Initialize string class that is automatically displayed as markdown by ``ipython``.

        Parameters
        ----------
        wrapped_str: str
            String to be wrapped.
        syntax: str
            Syntax highlighting which should be applied, by default None

        Note
        ----
        Possible syntax highlighting values can e.g. be found here:
        https://support.codebasehq.com/articles/tips-tricks/syntax-highlighting-in-markdown


        .. # noqa: DAR101
        """
        # This needs to be called data since ipython is looking for this attr
        self.data = str(wrapped_str)
        self.syntax = syntax

    def _repr_markdown_(self) -> str:
        """Render markdown automatically when in a ``ipython`` context.

        See:
        https://ipython.readthedocs.io/en/latest/config/integrating.html?highlight=_repr_markdown_#rich-display

        Returns
        -------
        str:
            Markdown string wrapped in a code block with syntax
            highlighting if syntax is not None.
        """
        if self.syntax is not None:
            return f"```{self.syntax}\n{self.data}\n```"
        else:
            return self.data

    def __str__(self) -> str:
        """Representation used by print and str."""
        return self._repr_markdown_()

    def __eq__(self, other: object) -> bool:
        """Equality check."""
        if isinstance(other, (str, MarkdownStr)):
            return str(self) == str(other)
        else:
            return NotImplemented


def display_file(path: StrOrPath, *, syntax: str | None = None) -> MarkdownStr:
    """Display a file with syntax highlighting ``syntax``.

    Parameters
    ----------
    path : StrOrPath
        Paths to the file
    syntax: str | None
        Syntax used for syntax highlighting. Defaults to None which means that the syntax is
        inferred based on the file extension. Pass the value ``""`` to deactivate syntax
        highlighting.

    Returns
    -------
    MarkdownStr
        File content with syntax highlighting to render in ipython.
    """
    if syntax is None:
        syntax = infer_file_format(path)
    with open(path, encoding="utf8") as file:
        return MarkdownStr(file.read(), syntax=syntax)



"""Package containing deprecated implementations which were removed.

To keep things organized the filenames should be like the
relative import path from glotaran root, but with ``_`` instead of ``.``.
E.g. ``glotaran.analysis.scheme`` would map to ``analysis_scheme.py``

The only exceptions to this rule are the root ``__init__.py`` which
is named ``glotaran_root.py`` and testing changed imports which should
be placed in ``test_changed_imports.py``.
"""



"""Test for ``glotaran.utils.tee``."""
from __future__ import annotations

from typing import TYPE_CHECKING

from glotaran.utils.tee import TeeContext

if TYPE_CHECKING:
    from _pytest.capture import CaptureFixture


def test_tee_context(capsys: CaptureFixture):
    """Test that print writes to stdout and is readable in context."""
    print_str = "foobar"
    expected = f"{print_str}\n"
    with TeeContext() as tee:
        print(print_str)
        result = tee.read()

    stdout, _ = capsys.readouterr()

    assert stdout == expected
    assert result == expected



"""Module containing the matrix provider classes."""
from __future__ import annotations

import warnings
from dataclasses import dataclass
from dataclasses import replace
from typing import TYPE_CHECKING
from typing import Any

import numpy as np
import xarray as xr

from glotaran.model import DatasetGroup
from glotaran.model import DatasetModel
from glotaran.model.dataset_model import has_dataset_model_global_model
from glotaran.model.dataset_model import iterate_dataset_model_global_megacomplexes
from glotaran.model.dataset_model import iterate_dataset_model_megacomplexes
from glotaran.model.interval_item import IntervalItem
from glotaran.model.item import fill_item
from glotaran.optimization.data_provider import DataProvider
from glotaran.optimization.data_provider import DataProviderLinked

if TYPE_CHECKING:
    from glotaran.typing.types import ArrayLike


@dataclass
class MatrixContainer:
    """A container of matrix and the corresponding clp labels."""

    clp_labels: list[str]
    """The clp labels."""
    matrix: np.ndarray
    """The matrix."""

    @property
    def is_index_dependent(self) -> bool:
        """Check if the matrix is index dependent.

        Returns
        -------
        bool
            Whether the matrix is index dependent.
        """
        return len(self.matrix.shape) == 3

    @staticmethod
    def apply_weight(matrix: ArrayLike, weight: ArrayLike) -> ArrayLike:
        """Apply weight on a matrix.

        Parameters
        ----------
        matrix : ArrayLike
            The matrix.
        weight : ArrayLike
            The weight.

        Returns
        -------
        ArrayLike
            The weighted matrix.
        """
        return (matrix.T * weight).T

    def create_weighted_matrix(self, weight: ArrayLike) -> MatrixContainer:
        """Create a matrix container with a weighted matrix.

        Parameters
        ----------
        weight : ArrayLike
            The weight.

        Returns
        -------
        MatrixContainer
            The weighted matrix.
        """
        return replace(self, matrix=self.apply_weight(self.matrix, weight))

    def create_scaled_matrix(self, scale: float) -> MatrixContainer:
        """Create a matrix container with a scaled matrix.

        Parameters
        ----------
        scale : float
            The scale.

        Returns
        -------
        MatrixContainer
            The scaled matrix.
        """
        return replace(self, matrix=self.matrix * scale)


class MatrixProvider:
    """A class to provide matrix calculations for optimization."""

    def __init__(self, dataset_group: DatasetGroup):
        """Initialize a matrix provider for a dataset group.

        Parameters
        ----------
        dataset_group : DatasetGroup
            The dataset group.
        """
        self._group = dataset_group
        self._matrix_containers: dict[str, MatrixContainer] = {}
        self._global_matrix_containers: dict[str, MatrixContainer] = {}
        self._data_provider: DataProvider

    @property
    def group(self) -> DatasetGroup:
        """Get the dataset group.

        Returns
        -------
        DatasetGroup
            The dataset group.
        """
        return self._group

    def get_matrix_container(self, dataset_label: str) -> MatrixContainer:
        """Get the matrix container for a dataset on an index on the global axis.

        Parameters
        ----------
        dataset_label : str
            The label of the dataset.

        Returns
        -------
        MatrixContainer
            The matrix container.
        """
        return self._matrix_containers[dataset_label]

    def calculate_dataset_matrices(self):
        """Calculate the matrices of the datasets in the dataset group."""
        for label, dataset_model in self.group.dataset_models.items():
            model_axis = self._data_provider.get_model_axis(label)
            global_axis = self._data_provider.get_global_axis(label)

            self._matrix_containers[label] = self.calculate_dataset_matrix(
                dataset_model, global_axis, model_axis
            )

    @staticmethod
    def calculate_dataset_matrix(
        dataset_model: DatasetModel,
        global_axis: ArrayLike,
        model_axis: ArrayLike,
        global_matrix: bool = False,
    ) -> MatrixContainer:
        """Calculate the matrix for a dataset on an index on the global axis.

        Parameters
        ----------
        dataset_model : DatasetModel
            The dataset model.
        global_axis: ArrayLike
            The global axis.
        model_axis: ArrayLike
            The model axis.
        global_matrix: bool
            Calculate the global megacomplexes if `True`.

        Returns
        -------
        MatrixContainer
            The resulting matrix container.
        """
        clp_labels: list[str] = []
        matrix = None

        megacomplex_iterator = iterate_dataset_model_megacomplexes(dataset_model)

        if global_matrix:
            megacomplex_iterator = iterate_dataset_model_global_megacomplexes(dataset_model)
            model_axis, global_axis = global_axis, model_axis

        for scale, megacomplex in megacomplex_iterator:
            this_clp_labels, this_matrix = megacomplex.calculate_matrix(  # type:ignore[union-attr]
                dataset_model, global_axis, model_axis
            )

            if scale is not None:
                this_matrix *= scale

            if matrix is None:
                clp_labels = this_clp_labels
                matrix = this_matrix
            else:
                clp_labels, matrix = MatrixProvider.combine_megacomplex_matrices(
                    matrix, this_matrix, clp_labels, this_clp_labels
                )
        return MatrixContainer(clp_labels, matrix)  # type:ignore[arg-type]

    @staticmethod
    def combine_megacomplex_matrices(
        matrix_left: ArrayLike,
        matrix_right: ArrayLike,
        clp_labels_left: list[str],
        clp_labels_right: list[str],
    ) -> tuple[list[str], ArrayLike]:
        """Calculate the matrix for a dataset on an index on the global axis.

        Parameters
        ----------
        matrix_left: ArrayLike
            The left matrix.
        matrix_right: ArrayLike
            The right matrix.
        clp_labels_left: list[str]
            The left clp labels.
        clp_labels_right: list[str]
            The right clp labels.

        Returns
        -------
        tuple[list[str], ArrayLike]:
            The combined clp labels and matrix.
        """
        result_clp_labels = clp_labels_left + [
            c for c in clp_labels_right if c not in clp_labels_left
        ]
        result_clp_size = len(result_clp_labels)

        if len(matrix_left.shape) < len(matrix_right.shape):
            matrix_left, matrix_right = matrix_right, matrix_left

        left_index_dependent = len(matrix_left.shape) == 3
        right_index_dependent = len(matrix_right.shape) == 3

        result_shape = (
            (matrix_left.shape[0], matrix_left.shape[1], result_clp_size)
            if left_index_dependent
            else (matrix_left.shape[0], result_clp_size)
        )

        result_matrix = np.zeros(result_shape, dtype=np.float64)
        for idx, label in enumerate(result_clp_labels):
            if label in clp_labels_left:
                if left_index_dependent:
                    result_matrix[:, :, idx] += matrix_left[:, :, clp_labels_left.index(label)]
                else:
                    result_matrix[:, idx] += matrix_left[:, clp_labels_left.index(label)]
            if label in clp_labels_right:
                if left_index_dependent:
                    result_matrix[:, :, idx] += (
                        matrix_right[:, :, clp_labels_right.index(label)]
                        if right_index_dependent
                        else matrix_right[:, clp_labels_right.index(label)]
                    )
                else:
                    result_matrix[:, idx] += matrix_right[:, clp_labels_right.index(label)]
        return result_clp_labels, result_matrix

    @staticmethod
    def does_interval_item_apply(prop: IntervalItem, index: int | None) -> bool:
        """Check if an interval item applies on an index.

        Parameters
        ----------
        prop : IntervalItem
            The interval property.
        index: int | None
            The index to check.

        Returns
        -------
        bool
            Whether the property applies.
        """
        if prop.has_interval() and index is None:
            warnings.warn(
                f"Interval property '{prop}' applies on a matrix which is "
                f"not index dependent. This will be an error in 0.9.0. Set "
                "'index_dependent: true' on the dataset model to fix the issue."
            )
            return True
        return prop.applies(index)

    def reduce_matrix(
        self,
        matrix: MatrixContainer,
        global_axis: ArrayLike,
    ) -> list[MatrixContainer]:
        """Reduce a matrix.

        Applies constraints and relations.

        Parameters
        ----------
        matrix : MatrixContainer
            The matrix.
        global_axis: ArrayLike
            The global axis.

        Returns
        -------
        MatrixContainer
            The resulting matrix container.
        """
        result = (
            [
                MatrixContainer(matrix.clp_labels, matrix.matrix[i, :, :])
                for i in range(global_axis.size)
            ]
            if matrix.is_index_dependent
            else [matrix] * global_axis.size
        )
        result = self.apply_relations(result, global_axis)
        result = self.apply_constraints(result, global_axis)
        return result

    def apply_constraints(
        self,
        matrices: list[MatrixContainer],
        global_axis: ArrayLike,
    ) -> list[MatrixContainer]:
        """Apply constraints on a matrix.

        Parameters
        ----------
        matrices: list[MatrixContainer],
            The matrices.
        global_axis: ArrayLike
            The global axis.

        Returns
        -------
        MatrixContainer
            The resulting matrix container.
        """
        model = self.group.model
        if len(model.clp_constraints) == 0:
            return matrices

        for i, index in enumerate(global_axis):
            matrix = matrices[i]
            clp_labels = matrix.clp_labels
            removed_clp_labels = [
                c.target
                for c in model.clp_constraints
                if c.target in clp_labels and self.does_interval_item_apply(c, index)
            ]
            if len(removed_clp_labels) == 0:
                continue
            reduced_clp_labels = [c for c in clp_labels if c not in removed_clp_labels]
            mask = [label in reduced_clp_labels for label in clp_labels]
            reduced_matrix = matrix.matrix[:, mask]
            matrices[i] = MatrixContainer(reduced_clp_labels, reduced_matrix)
        return matrices

    def apply_relations(
        self,
        matrices: list[MatrixContainer],
        global_axis: ArrayLike,
    ) -> list[MatrixContainer]:
        """Apply relations on a matrix.

        Parameters
        ----------
        matrices: list[MatrixContainer],
            The matrices.
        global_axis: ArrayLike
            The global axis.

        Returns
        -------
        MatrixContainer
            The resulting matrix container.
        """
        model = self.group.model
        parameters = self.group.parameters

        if len(model.clp_relations) == 0:
            return matrices

        for i, index in enumerate(global_axis):
            matrix = matrices[i]

            clp_labels = matrix.clp_labels
            relation_matrix = np.diagflat([1.0 for _ in clp_labels])

            idx_to_delete = []
            for relation in model.clp_relations:
                if relation.target in clp_labels and self.does_interval_item_apply(
                    relation, index
                ):
                    if relation.source not in clp_labels:
                        continue

                    relation = fill_item(relation, model, parameters)  # type:ignore[arg-type]
                    source_idx = clp_labels.index(relation.source)
                    target_idx = clp_labels.index(relation.target)
                    relation_matrix[target_idx, source_idx] = relation.parameter
                    idx_to_delete.append(target_idx)

            if len(idx_to_delete) == 0:
                continue

            reduced_clp_labels = [
                label for i, label in enumerate(clp_labels) if i not in idx_to_delete
            ]
            relation_matrix = np.delete(relation_matrix, idx_to_delete, axis=1)
            reduced_matrix = matrix.matrix @ relation_matrix
            matrices[i] = MatrixContainer(reduced_clp_labels, reduced_matrix)
        return matrices

    def get_result(self) -> tuple[dict[str, xr.DataArray], dict[str, xr.DataArray]]:
        """Get the results of the matrix calculations.

        Returns
        -------
        tuple[dict[str, xr.DataArray], dict[str, xr.DataArray]]
            A tuple of the matrices and global matrices.

        .. # noqa: DAR202
        .. # noqa: DAR401
        """
        matrices = {}
        global_matrices = {}
        for label, matrix_container in self._matrix_containers.items():
            model_dimension = self._data_provider.get_model_dimension(label)
            model_axis = self._data_provider.get_model_axis(label)
            matrix_coords: tuple[tuple[str, Any], tuple[str, Any], tuple[str, list[str]]] | tuple[
                tuple[str, Any], tuple[str, list[str]]
            ] = (
                (model_dimension, model_axis),
                ("clp_label", matrix_container.clp_labels),
            )
            if matrix_container.is_index_dependent:
                global_dimension = self._data_provider.get_global_dimension(label)
                global_axis = self._data_provider.get_global_axis(label)
                matrix_coords = (
                    (global_dimension, global_axis),
                    matrix_coords[0],
                    matrix_coords[1],
                )
            matrices[label] = xr.DataArray(matrix_container.matrix, coords=matrix_coords)

        for label, matrix_container in self._global_matrix_containers.items():
            global_dimension = self._data_provider.get_global_dimension(label)
            global_axis = self._data_provider.get_global_axis(label)
            global_matrices[label] = xr.DataArray(
                matrix_container.matrix,
                coords=(
                    (global_dimension, global_axis),
                    ("global_clp_label", matrix_container.clp_labels),
                ),
            )

        return global_matrices, matrices

    def calculate(self):
        """Calculate the matrices for optimization.

        .. # noqa: DAR401
        """
        raise NotImplementedError

    @property
    def number_of_clps(self) -> int:
        """Return number of conditionally linear parameters.

        Raises
        ------
        NotImplementedError
            This property needs to be implemented by subclasses.

        See Also
        --------
        MatrixProviderUnlinked
        MatrixProviderLinked
        """
        raise NotImplementedError


class MatrixProviderUnlinked(MatrixProvider):
    """A class to provide matrix calculations for optimization of unlinked dataset groups."""

    def __init__(self, group: DatasetGroup, data_provider: DataProvider):
        """Initialize a matrix provider for an unlinked dataset group.

        Parameters
        ----------
        dataset_group : DatasetGroup
            The dataset group.
        data_provider : DataProvider
            The data provider.
        """
        super().__init__(group)
        self._data_provider = data_provider
        self._prepared_matrix_container: dict[str, list[MatrixContainer]] = {}
        self._full_matrices: dict[str, ArrayLike] = {}

    def get_global_matrix_container(self, dataset_label: str) -> MatrixContainer:
        """Get the global matrix container for a dataset.

        Parameters
        ----------
        dataset_label : str
            The label of the dataset.

        Returns
        -------
        MatrixContainer
            The matrix container.
        """
        return self._global_matrix_containers[dataset_label]

    def get_prepared_matrix_container(
        self, dataset_label: str, global_index: int
    ) -> MatrixContainer:
        """Get the prepared matrix container for a dataset on an index on the global axis.

        Parameters
        ----------
        dataset_label : str
            The label of the dataset.
        global_index : int
            The index on the global axis.

        Returns
        -------
        MatrixContainer
            The matrix container.
        """
        return self._prepared_matrix_container[dataset_label][global_index]

    def get_full_matrix(self, dataset_label: str) -> ArrayLike:
        """Get the full matrix of a dataset.

        Parameters
        ----------
        dataset_label : str
            The label of the dataset.

        Returns
        -------
        ArrayLike
            The matrix.
        """
        return self._full_matrices[dataset_label]

    def calculate(self):
        """Calculate the matrices for optimization."""
        self.calculate_dataset_matrices()
        self.calculate_global_matrices()
        self.calculate_prepared_matrices()
        self.calculate_full_matrices()

    def calculate_global_matrices(self):
        """Calculate the global matrices of the datasets in the dataset group."""
        for label, dataset_model in self.group.dataset_models.items():
            if has_dataset_model_global_model(dataset_model):
                model_axis = self._data_provider.get_model_axis(label)
                global_axis = self._data_provider.get_global_axis(label)
                self._global_matrix_containers[label] = self.calculate_dataset_matrix(
                    dataset_model, global_axis, model_axis, global_matrix=True
                )

    def calculate_prepared_matrices(self):
        """Calculate the prepared matrices of the datasets in the dataset group."""
        for label, dataset_model in self.group.dataset_models.items():
            if has_dataset_model_global_model(dataset_model):
                continue
            scale = float(dataset_model.scale or 1)
            weight = self._data_provider.get_weight(label)
            self._prepared_matrix_container[label] = self.reduce_matrix(
                self.get_matrix_container(label).create_scaled_matrix(scale),
                self._data_provider.get_global_axis(label),
            )

            if weight is not None:
                self._prepared_matrix_container[label] = [
                    matrix.create_weighted_matrix(weight[:, i])
                    for i, matrix in enumerate(self._prepared_matrix_container[label])
                ]

    def calculate_full_matrices(self):
        """Calculate the full matrices of the datasets in the dataset group."""
        for label, dataset_model in self.group.dataset_models.items():
            if has_dataset_model_global_model(dataset_model):
                global_matrix_container = self.get_global_matrix_container(label)
                global_matrix = global_matrix_container.matrix
                matrix_container = self.get_matrix_container(label)
                matrix = matrix_container.matrix

                if matrix_container.is_index_dependent:
                    full_matrix = np.concatenate(
                        [
                            np.kron(global_matrix[i, :], matrix[i, :, :])
                            for i in range(matrix.shape[0])
                        ]
                    )
                else:
                    full_matrix = np.kron(global_matrix, matrix)

                weight = self._data_provider.get_flattened_weight(label)
                if weight is not None:
                    full_matrix = MatrixContainer.apply_weight(full_matrix, weight)

                self._full_matrices[label] = full_matrix

    @property
    def number_of_clps(self) -> int:
        """Return number of conditionally linear parameters.

        Returns
        -------
        int
        """
        nr_of_clps = 0
        for dataset_label, dataset_model in self.group.dataset_models.items():
            if has_dataset_model_global_model(dataset_model):
                model_clp_labels = self.get_matrix_container(dataset_label).clp_labels
                global_clp_labels = self.get_global_matrix_container(dataset_label).clp_labels
                nr_of_clps += len(model_clp_labels) * len(global_clp_labels)
            else:
                global_axis_indexes = range(
                    len(self._data_provider.get_global_axis(dataset_label))
                )
                nr_of_clps += sum(
                    len(self.get_prepared_matrix_container(dataset_label, index).clp_labels)
                    for index in global_axis_indexes
                )

        return nr_of_clps


class MatrixProviderLinked(MatrixProvider):
    """A class to provide matrix calculations for optimization of linked dataset groups."""

    def __init__(self, group: DatasetGroup, data_provider: DataProviderLinked):
        """Initialize a matrix provider for a linked dataset group.

        Parameters
        ----------
        dataset_group : DatasetGroup
            The dataset group.
        data_provider : DataProviderLinked
            The data provider.
        """
        super().__init__(group)
        self._data_provider: DataProviderLinked = data_provider
        self._aligned_full_clp_labels: list[list[str]] = [
            None  # type:ignore[list-item]
        ] * self._data_provider.aligned_global_axis.size
        self._aligned_matrices: list[MatrixContainer] = [
            None  # type:ignore[list-item]
        ] * self._data_provider.aligned_global_axis.size

    @property
    def aligned_full_clp_labels(self) -> list[list[str]]:
        """Get the aligned full clp labels.

        Returns
        -------
        list[list[str]]
            The full aligned clp labels.
        """
        return self._aligned_full_clp_labels

    def get_aligned_matrix_container(self, global_index: int) -> MatrixContainer:
        """Get the aligned matrix container for an index on the aligned global axis.

        Parameters
        ----------
        global_index : int
            The index on the global axis.

        Returns
        -------
        MatrixContainer
            The matrix container.
        """
        return self._aligned_matrices[global_index]

    def calculate(self):
        """Calculate the matrices for optimization."""
        self.calculate_dataset_matrices()
        self.calculate_aligned_matrices()

    def calculate_aligned_matrices(self):
        """Calculate the aligned matrices of the dataset group."""
        full_clp_labels = self.align_full_clp_labels()
        for i, global_index_value in enumerate(self._data_provider.aligned_global_axis):
            matrix_containers = []
            group_label = self._data_provider.get_aligned_group_label(i)
            for label, index in zip(
                self._data_provider.group_definitions[group_label],
                self._data_provider.get_aligned_dataset_indices(i),
            ):
                matrix_container_temp = self._matrix_containers[label]
                if matrix_container_temp.is_index_dependent:
                    matrix_containers.append(
                        MatrixContainer(
                            clp_labels=matrix_container_temp.clp_labels,
                            matrix=matrix_container_temp.matrix[index],
                        )
                    )
                else:
                    matrix_containers.append(matrix_container_temp)

            matrix_scales = [
                self.group.dataset_models[label].scale
                if self.group.dataset_models[label].scale is not None
                else 1
                for label in self._data_provider.group_definitions[group_label]
            ]

            group_matrix = self.align_matrices(
                matrix_containers, matrix_scales  # type:ignore[arg-type]
            )

            self._aligned_full_clp_labels[i] = full_clp_labels[group_label]
            group_matrix_single = self.reduce_matrix(
                group_matrix, np.array([self._data_provider.aligned_global_axis[i]])
            )[0]

            weight = self._data_provider.get_aligned_weight(i)
            if weight is not None:
                group_matrix_single = group_matrix_single.create_weighted_matrix(weight)

            self._aligned_matrices[i] = group_matrix_single

    def align_full_clp_labels(self) -> dict[str, list[str]]:
        """Align the unreduced clp labels.

        Returns
        -------
        dict[str, list[str]]
            The aligned clp for every group.
        """
        aligned_full_clp_labels: dict[str, list[str]] = {}

        for (
            group_label,
            dataset_labels,
        ) in self._data_provider.group_definitions.items():
            aligned_full_clp_labels[group_label] = []
            for dataset_label in dataset_labels:
                aligned_full_clp_labels[group_label] += [
                    label
                    for label in self.get_matrix_container(dataset_label).clp_labels
                    if label not in aligned_full_clp_labels[group_label]
                ]
        return aligned_full_clp_labels

    @staticmethod
    def align_matrices(matrices: list[MatrixContainer], scales: list[float]) -> MatrixContainer:
        """Align matrices.

        Parameters
        ----------
        matrices : list[MatrixContainer]
            The matrices to align.
        scales : list[float]
            The scales of the matrices.

        Returns
        -------
        MatrixContainer
            The aligned matrix container.
        """
        if len(matrices) == 1:
            return matrices[0]
        masks = []
        full_clp_labels: list[str] = []
        sizes = []
        dim1 = 0
        for matrix in matrices:
            clp_labels = matrix.clp_labels
            model_axis_size = matrix.matrix.shape[0]
            sizes.append(model_axis_size)
            dim1 += model_axis_size
            if len(full_clp_labels) == 0:
                full_clp_labels = clp_labels.copy()
                masks.append([i for i, _ in enumerate(clp_labels)])
            else:
                mask = []
                for c in clp_labels:
                    if c not in full_clp_labels:
                        full_clp_labels.append(c)
                    mask.append(full_clp_labels.index(c))
                masks.append(mask)
        dim2 = len(full_clp_labels)
        full_matrix = np.zeros((dim1, dim2), dtype=np.float64)
        start = 0
        for i, m in enumerate(matrices):
            end = start + sizes[i]
            full_matrix[start:end, masks[i]] = m.matrix * scales[i]
            start = end

        return MatrixContainer(full_clp_labels, full_matrix)

    @property
    def number_of_clps(self) -> int:
        """Return number of conditionally linear parameters.

        Returns
        -------
        int
        """
        global_axis_indexes = range(len(self._data_provider.aligned_global_axis))
        return sum(
            len(self.get_aligned_matrix_container(index).clp_labels)
            for index in global_axis_indexes
        )



"""Deprecated functionality export for 'glotaran.examples.sequential'."""

sequential_deprecation_mapping = {
    "sim_model": "SIMULATION_MODEL",
    "dataset": "DATASET",
    "model": "MODEL",
    "scheme": "SCHEME",
}
shared_deprecation_mapping = {
    "wanted_parameter": "SIMULATION_PARAMETERS",
    "parameter": "PARAMETERS",
    "_time": "TIME_AXIS",
    "_spectral": "SPECTRAL_AXIS",
}


def __getattr__(attribute_name: str):
    from glotaran.deprecation.deprecation_utils import deprecate_module_attribute

    for deprecated, new in shared_deprecation_mapping.items():
        if attribute_name == deprecated:
            return deprecate_module_attribute(
                deprecated_qual_name=f"glotaran.examples.sequential.{deprecated}",
                new_qual_name=f"glotaran.testing.simulated_data.shared_decay.{new}",
                to_be_removed_in_version="0.8.0",
            )

    for deprecated, new in sequential_deprecation_mapping.items():
        if attribute_name == deprecated:
            return deprecate_module_attribute(
                deprecated_qual_name=f"glotaran.examples.sequential.{deprecated}",
                new_qual_name=f"glotaran.testing.simulated_data.sequential_spectral_decay.{new}",
                to_be_removed_in_version="0.8.0",
            )

    raise AttributeError(f"module {__name__} has no attribute {attribute_name}")



"""Module containing the NetCDF4 Data IO plugin."""
from __future__ import annotations

# Needed to prevent a netCDF4 RuntimeWarning at import time
# Ref.: https://github.com/pydata/xarray/issues/7259
import netCDF4  # noqa: F401
import xarray as xr

from glotaran.io import DataIoInterface
from glotaran.io import register_data_io


@register_data_io("nc")
class NetCDFDataIo(DataIoInterface):
    """Plugin for NetCDF4 data io."""

    def load_dataset(self, file_name: str) -> xr.Dataset | xr.DataArray:
        """Load a ``*.nc`` file into a :xarraydoc:`Dataset` or :xarraydoc:`DataArray`.

        Parameters
        ----------
        file_name: str
            Path to the ``*.nc`` file that should be loaded.

        Returns
        -------
        xr.Dataset | xr.DataArray
        """
        with xr.open_dataset(file_name) as ds:
            return ds.load()

    def save_dataset(
        self,
        dataset: xr.Dataset,
        file_name: str,
        *,
        data_filters: list[str] | None = None,
    ):
        """Write a :xarraydoc:`Dataset` to the ``*.nc`` at path ``file_name``.

        Parameters
        ----------
        dataset: xr.Dataset
            :xarraydoc:`Dataset` that should be written to file.
        file_name: str
            Path of the file to write ``dataset`` to.
        data_filters: list[str] | None
            List of data variable names that should be written to file. Defaults to None.
        """
        data_to_save = dataset if data_filters is None else dataset[data_filters]
        data_to_save.to_netcdf(file_name, mode="w")



import numpy as np
import pandas as pd
import pytest
import xarray as xr

from glotaran.builtin.io.sdt.sdt_file_reader import SdtDataIo
from glotaran.builtin.io.sdt.test import TEMPORAL_DATA


@pytest.mark.parametrize(
    "test_file_path, result_file_path, index",
    [
        (TEMPORAL_DATA["sdt"], TEMPORAL_DATA["csv"], [1]),
        (TEMPORAL_DATA["sdt"], TEMPORAL_DATA["csv"], None),
    ],
)
@pytest.mark.filterwarnings("ignore:There was no `index`:UserWarning")
def test_read_sdt(test_file_path, result_file_path, index):
    sdt_reader = SdtDataIo("sdt")
    test_dataset = sdt_reader.load_dataset(test_file_path, index=index)
    result_df = pd.read_csv(
        result_file_path, skiprows=1, sep=r"\s+", dtype={"Delay": float, "Data": np.uint16}
    )
    result_df.Delay = result_df.Delay * 1e-9
    result_traces = pd.DataFrame([result_df.Data.values], columns=result_df.Delay)
    assert isinstance(test_dataset, xr.Dataset)
    assert np.all(test_dataset.data.T[0] == result_df.Data.values)

    assert test_dataset.data.T.shape == result_traces.values.shape
    assert np.allclose(test_dataset.time, np.array(result_traces.columns))



"""This package contains the initial concentration item."""
from __future__ import annotations

import numpy as np

from glotaran.model import ModelItem
from glotaran.model import ParameterType
from glotaran.model import item


@item
class InitialConcentration(ModelItem):
    """An initial concentration describes the population of the compartments at
    the beginning of an experiment."""

    compartments: list[str]
    parameters: list[ParameterType]
    exclude_from_normalize: list[str] = []

    def normalized(self) -> np.ndarray:
        normalized = np.array(self.parameters)
        idx = [c not in self.exclude_from_normalize for c in self.compartments]
        normalized[idx] /= np.sum(normalized[idx])
        return normalized



"""Module containing general helper functions."""

from typing import Any

import numpy as np


def nan_or_equal(lhs: Any, rhs: Any) -> bool:
    """Compare values which can be nan for equality.

    This helper function is needed because ``np.nan == np.nan`` returns ``False``.

    Parameters
    ----------
    lhs: Any
        Left hand side value.
    rhs: Any
        Right hand side value.

    Returns
    -------
    bool
        Whether or not values are equal.
    """
    if isinstance(lhs, (int, float)) and isinstance(rhs, (int, float)):
        return (np.isnan(lhs) and np.isnan(rhs)) or lhs == rhs
    return lhs == rhs



"""The glotaran result registry module."""
from __future__ import annotations

import re
from pathlib import Path
from warnings import warn

from glotaran.io import load_result
from glotaran.io import save_result
from glotaran.project.project_registry import ProjectRegistry
from glotaran.project.result import Result


class ProjectResultRegistry(ProjectRegistry):
    """A registry for results."""

    result_pattern = re.compile(r".+_run_\d{4}$")

    def __init__(self, directory: Path):
        """Initialize a result registry.

        Parameters
        ----------
        directory : Path
            The registry directory.
        """
        super().__init__(
            directory / "results",
            [],
            lambda path: load_result(path / "result.yml", format_name="yml"),
            item_name="Result",
        )

    def is_item(self, path: Path) -> bool:
        """Check if the path contains an registry item.

        Parameters
        ----------
        path : Path
            The path to check.

        Returns
        -------
        bool :
            Whether the path contains an item.
        """
        return path.is_dir()

    def previous_result_paths(self, base_name: str) -> list[Path]:
        """List previous result paths with base_name.

        Parameters
        ----------
        base_name: str
            The base name for the result provided by user or derived from model name.

        Returns
        -------
        list[Path]
            Paths to previous results with name ``base_name``.
        """
        return sorted(self.directory.glob(f"{base_name}_run_*"))

    def _latest_result_path_fallback(self, name: str, *, latest: bool = False) -> Path:
        """Fallback when a user forgets to specify the run to get a result.

        If ``name`` contains the run number this will just return ``name``,
        else we try to get the name of the latest run.

        Parameters
        ----------
        name: str
            Name of the result, which should contain the run specifier.
        latest: bool
            Flag to deactivate warning about using latest result. Defaults to False.


        Returns
        -------
        Path
            Path to the result (latest result if ``name`` does not match the result pattern).

        Raises
        ------
        ValueError
            Raised if result does not exist.
        """
        if re.match(self.result_pattern, name) is None:
            if latest is False:
                warn(
                    UserWarning(
                        f"Result name {name!r} is missing the run specifier, "
                        "falling back to try getting latest result. "
                        "Use latest=True to mute this warning."
                    ),
                    stacklevel=3,
                )
            previous_result_paths = self.previous_result_paths(name) or [Path(name)]
            name = previous_result_paths[-1].stem
        path = self._directory / name
        if self.is_item(path):
            return path

        raise ValueError(
            f"Result {name!r} does not exist. Known Results are: {list(self.items.keys())}"
        )

    def create_result_run_name(self, base_name: str) -> str:
        """Create a result name for a model.

        Parameters
        ----------
        base_name: str
            The base name for the result provided by user or derived from model name.

        Returns
        -------
        str :
            Folder name for the new result to be saved in.
        """
        previous_results = self.previous_result_paths(base_name)
        if not previous_results:
            return f"{base_name}_run_0000"
        latest_result_run_nr = int(previous_results[-1].stem.replace(f"{base_name}_run_", ""))
        return f"{base_name}_run_{latest_result_run_nr+1:04}"

    def save(self, name: str, result: Result):
        """Save a result.

        Parameters
        ----------
        name : str
            The name of the result.
        result : Result
            The result to save.
        """
        run_name = self.create_result_run_name(name)
        result_path = self.directory / run_name / "result.yml"
        save_result(result, result_path, format_name="yml")



"""Module containing the SDT Data IO plugin."""
from __future__ import annotations

import warnings

import numpy as np
import xarray as xr
from sdtfile import SdtFile

from glotaran.io import DataIoInterface
from glotaran.io import register_data_io
from glotaran.io.prepare_dataset import prepare_time_trace_dataset


@register_data_io("sdt")
class SdtDataIo(DataIoInterface):
    """Plugin for SDT data io."""

    def load_dataset(
        self,
        file_name: str,
        *,
        index: np.ndarray | None = None,
        flim: bool = False,
        dataset_index: int | None = None,
        swap_axis: bool = False,
        orig_time_axis_index: int = 2,
    ) -> xr.Dataset:
        """Read a `*.sdt` file and returns a :xarraydoc:`Dataset` containing its data.

        Parameters
        ----------
        file_name: str
            Path to the sdt file which should be read.

        index: np.ndarray | None
            This is only needed if `type_of_data=="st"`, since `*.sdt` files,
            which only contain spectral temporal data, lack the spectral information.
            Thus for the spectral axis data need to be given by the user.

        flim: bool
            Set true if reading a result from a FLIM measurement. Defaults to False.

        dataset_index: int
            If the `*.sdt` file contains multiple datasets the index will used
            to select the wanted one. Defaults to 0.

        swap_axis: bool
            Flag to switch a wavelength explicit `input_df` to time explicit `input_df`,
            before generating the SpectralTemporalDataset. Defaults to False.

        orig_time_axis_index: int
            Index of the axis which corresponds to the time axis.
            I.e. for data of shape (64, 64, 256), which are a 64x64 pixel map
            with 256 time steps, orig_time_axis_index=2.

        Returns
        -------
        xr.Dataset

        Raises
        ------
        IndexError
            If the length of the index array is incompatible with the data.
        """
        sdt_parser = SdtFile(file_name)
        if not dataset_index:
            # looking at the source code of SdtFile, times and data
            # always have the same len, so only one needs to be checked
            nr_of_datasets = len(sdt_parser.times)
            if nr_of_datasets > 1:
                warnings.warn(
                    UserWarning(
                        f"The file '{file_name}' contains {nr_of_datasets} Datasets.\n "
                        f"By default only the first Dataset will be read. "
                        f"If you only need the first Dataset and want get rid of "
                        f"this warning you can set dataset_index=0."
                    ),
                    stacklevel=4,
                )
            dataset_index = 0
        times: np.ndarray = sdt_parser.times[dataset_index]
        raw_data: np.ndarray = sdt_parser.data[dataset_index]

        if index and len(index) != raw_data.shape[0]:
            raise IndexError(
                f"The Dataset contains {raw_data.shape[0]} measurements, but the "
                f"indices supplied are {len(index)}."
            )
        elif not index and not flim:
            warnings.warn(
                UserWarning(
                    f"There was no `index` provided."
                    f"That for the indices will be a entry count(integers)."
                    f"To prevent this warning from being shown, provide "
                    f"a list of indices, with len(index)={raw_data.shape[0]}"
                ),
                stacklevel=4,
            )

        if flim:
            if orig_time_axis_index != 2:
                np.swapaxes(raw_data, 2, orig_time_axis_index)

            full_data = xr.DataArray(raw_data, coords={"time": times}, dims=["x", "y", "time"])
            data = full_data.stack(pixel=("x", "y")).to_dataset(name="data")
            data["full_data"] = full_data.rename({"x": "pixel_x", "y": "pixel_y"})
            data["data_intensity_map"] = (
                data.data.groupby("pixel").sum().unstack().rename({"x": "pixel_x", "y": "pixel_y"})
            )
        else:
            if swap_axis:
                raw_data = raw_data.T
            if not index:
                index = np.arange(raw_data.shape[0])
            data = xr.DataArray(raw_data.T, coords=[("time", times), ("spectral", index)])
            data = prepare_time_trace_dataset(data)
        return data



"""Module containing the optimize function."""
from __future__ import annotations

from glotaran.optimization.optimizer import Optimizer
from glotaran.project import Result
from glotaran.project import Scheme


def optimize(scheme: Scheme, verbose: bool = True, raise_exception: bool = False) -> Result:
    """Optimize a scheme.

    Parameters
    ----------
    scheme : Scheme
        The optimization scheme.
    verbose : bool
        Deactivate printing of logs if `False`.
    raise_exception : bool
        Raise exceptions during optimizations instead of gracefully exiting if `True`.

    Returns
    -------
    Result
        The result of the optimization.
    """
    optimizer = Optimizer(scheme, verbose, raise_exception)
    optimizer.optimize()
    return optimizer.create_result()



import click

from glotaran import __version__ as VERSION
from glotaran.cli.commands.optimize import optimize_cmd
from glotaran.cli.commands.pluginlist import plugin_list_cmd
from glotaran.cli.commands.print import print_cmd
from glotaran.cli.commands.validate import validate_cmd


class Cli(click.Group):
    """The glotaran CLI implementation of :class:`click.group`"""

    def __init__(self, *args, **kwargs):
        click.echo(
            "[DEPRECATED] The pyglotaran command line interface will be removed without "
            "replacement in version '0.8.0' since it lacks a lot of essential functionality and "
            "isn't used. The currently recommended usage of pyglotaran is in JuPyteR notebooks.",
            err=True,
        )
        self.help_priorities = {}
        super().__init__(*args, **kwargs)

    def get_help(self, ctx):
        self.list_commands = self.list_commands_for_help
        return super().get_help(ctx)

    def list_commands_for_help(self, ctx):
        """reorder the list of commands when listing the help"""
        commands = super().list_commands(ctx)
        return (
            c[1]
            for c in sorted(
                (self.help_priorities.get(command, 1), command) for command in commands
            )
        )

    def command(self, *args, **kwargs):
        """Behaves the same as `click.Group.command()` except capture
        a priority for listing command names in help.
        """
        help_priority = kwargs.pop("help_priority", 1)
        help_priorities = self.help_priorities

        def decorator(f):
            cmd = super(Cli, self).command(*args, **kwargs)(f)
            help_priorities[cmd.name] = help_priority
            return cmd

        return decorator


@click.group(cls=Cli)
@click.version_option(version=VERSION)
def main(prog_name="glotaran"):
    """The glotaran CLI main function."""


main.add_command(
    main.command(
        name="pluginlist", short_help="Prints a list of installed plugins.", help_priority=4
    )(plugin_list_cmd)
)
main.add_command(
    main.command(name="print", short_help="Prints a model as markdown.", help_priority=3)(
        print_cmd
    )
)
main.add_command(
    main.command(name="validate", short_help="Validates a model file.", help_priority=2)(
        validate_cmd
    )
)
main.add_command(
    main.command(name="optimize", short_help="Optimizes a model.", help_priority=1)(optimize_cmd)
)


if __name__ == "__main__":
    raise SystemExit(main(prog_name="glotaran"))



import warnings
from textwrap import dedent

import numpy as np
import pytest
from attrs import evolve

from glotaran.io import load_model
from glotaran.io import load_parameters
from glotaran.model import fill_item
from glotaran.optimization.optimize import optimize
from glotaran.project import Scheme
from glotaran.simulation import simulate

MODEL_BASE = """\
default_megacomplex: decay
dataset:
    dataset1:
        megacomplex: [mc1]
        initial_concentration: j1
        irf: irf1
initial_concentration:
    j1:
        compartments: [s1]
        parameters: [j.1]
megacomplex:
    mc1:
        k_matrix: [k1]
    mc2:
        type: spectral
        shape:
            s1: sh1
k_matrix:
    k1:
        matrix:
            (s1, s1): kinetic.1
shape:
    sh1:
        type: one
"""
MODEL_NO_IRF_DISPERSION = f"""\
{MODEL_BASE}
irf:
    irf1:
        type: gaussian
        center: irf.center
        width: irf.width
"""
MODEL_SIMPLE_IRF_DISPERSION = f"""\
{MODEL_BASE}
irf:
    irf1:
        type: spectral-gaussian
        center: irf.center
        width: irf.width
        dispersion_center: irf.dispersion_center
        center_dispersion_coefficients: [irf.cdc1]
"""
MODEL_MULTI_IRF_DISPERSION = f"""\
{MODEL_BASE}
irf:
    irf1:
        type: spectral-multi-gaussian
        center: [irf.center]
        width: [irf.width]
        dispersion_center: irf.dispersion_center
        center_dispersion_coefficients: [irf.cdc1, irf.cdc2]
        width_dispersion_coefficients: [irf.wdc1]
"""

MODEL_MULTIPULSE_IRF_DISPERSION = f"""\
{MODEL_BASE}
irf:
    irf1:
        type: spectral-multi-gaussian
        center: [irf.center1, irf.center2]
        width: [irf.width]
        dispersion_center: irf.dispersion_center
        center_dispersion_coefficients: [irf.cdc1, irf.cdc2, irf.cdc3]
"""

PARAMETERS_BASE = """\
j:
    - ['1', 1, {'vary': False, 'non-negative': False}]
kinetic:
    - ['1', 0.5, {'non-negative': False}]
"""

PARAMETERS_NO_IRF_DISPERSION = f"""\
{PARAMETERS_BASE}
irf:
    - ['center', 0.3]
    - ['width', 0.1]
"""

PARAMETERS_SIMPLE_IRF_DISPERSION = f"""\
{PARAMETERS_BASE}
irf:
    - ['center', 0.3]
    - ['width', 0.1]
    - ['dispersion_center', 400, {{'vary': False}}]
    - ['cdc1', 0.5]
"""

# What is this?
PARAMETERS_MULTI_IRF_DISPERSION = f"""\
{PARAMETERS_BASE}
irf:
    - ["center", 0.3]
    - ["width", 0.1]
    - ["dispersion_center", 400, {{"vary": False}}]
    - ["cdc1", 0.1]
    - ["cdc2", 0.01]
    - ["wdc1", 0.025]
"""

PARAMETERS_MULTIPULSE_IRF_DISPERSION = f"""\
{PARAMETERS_BASE}
irf:
    - ["center1", 0.3]
    - ["center2", 0.4]
    - ['width', 0.1]
    - ['dispersion_center', 400, {{'vary': False}}]
    - ["cdc1", 0.5]
    - ["cdc2", 0.1]
    - ["cdc3", -0.01]
"""


def _time_axis():
    time_p1 = np.linspace(-1, 1, 20, endpoint=False)
    time_p2 = np.linspace(1, 2, 10, endpoint=False)
    time_p3 = np.geomspace(2, 20, num=20)
    return np.array(np.concatenate([time_p1, time_p2, time_p3]))


def _spectral_axis():
    return np.linspace(300, 500, 3)


def _calculate_irf_position(
    index, center, dispersion_center=None, center_dispersion_coefficients=None
):
    if center_dispersion_coefficients is None:
        center_dispersion_coefficients = []
    if dispersion_center is not None:
        distance = (index - dispersion_center) / 100
        for i, coefficient in enumerate(center_dispersion_coefficients):
            center += coefficient * np.power(distance, i + 1)
    return center


class NoIrfDispersion:
    model = load_model(MODEL_NO_IRF_DISPERSION, format_name="yml_str")
    parameters = load_parameters(PARAMETERS_NO_IRF_DISPERSION, format_name="yml_str")
    axis = {"time": _time_axis(), "spectral": _spectral_axis()}


class SimpleIrfDispersion:
    model = load_model(MODEL_SIMPLE_IRF_DISPERSION, format_name="yml_str")
    parameters = load_parameters(PARAMETERS_SIMPLE_IRF_DISPERSION, format_name="yml_str")
    axis = {"time": _time_axis(), "spectral": _spectral_axis()}


class MultiIrfDispersion:
    model = load_model(MODEL_MULTI_IRF_DISPERSION, format_name="yml_str")
    parameters = load_parameters(PARAMETERS_MULTI_IRF_DISPERSION, format_name="yml_str")
    axis = {"time": _time_axis(), "spectral": _spectral_axis()}


class MultiCenterIrfDispersion:
    model = load_model(MODEL_MULTIPULSE_IRF_DISPERSION, format_name="yml_str")
    parameters = load_parameters(PARAMETERS_MULTIPULSE_IRF_DISPERSION, format_name="yml_str")
    axis = {"time": _time_axis(), "spectral": _spectral_axis()}


@pytest.mark.parametrize(
    "suite",
    [
        NoIrfDispersion,
        SimpleIrfDispersion,
        MultiIrfDispersion,
        MultiCenterIrfDispersion,
    ],
)
def test_spectral_irf(suite):
    model = suite.model
    assert model.valid(), model.validate()

    parameters = suite.parameters
    assert model.valid(parameters), model.validate(parameters)

    sim_model = evolve(model)
    sim_model.dataset["dataset1"].global_megacomplex = ["mc2"]
    print(sim_model)
    dataset = simulate(sim_model, "dataset1", parameters, suite.axis)

    assert dataset.data.shape == (suite.axis["time"].size, suite.axis["spectral"].size)

    data = {"dataset1": dataset}

    scheme = Scheme(
        model=model,
        parameters=parameters,
        data=data,
        maximum_number_function_evaluations=20,
    )
    result = optimize(scheme)

    for param in result.optimized_parameters.all():
        assert np.allclose(param.value, parameters.get(param.label).value, rtol=1e-1)

    resultdata = result.data["dataset1"]

    # print(resultdata)
    assert np.array_equal(dataset["time"], resultdata["time"])
    assert np.array_equal(dataset["spectral"], resultdata["spectral"])
    assert dataset.data.shape == resultdata.data.shape
    assert dataset.data.shape == resultdata.fitted_data.shape
    # assert np.allclose(dataset.data, resultdata.fitted_data, atol=1e-14)

    fit_data_max_at_start = resultdata.fitted_data.isel(spectral=0).argmax(axis=0)
    fit_data_max_at_end = resultdata.fitted_data.isel(spectral=-1).argmax(axis=0)

    if suite is NoIrfDispersion:
        assert "center_dispersion_1" not in resultdata
        assert fit_data_max_at_start == fit_data_max_at_end
    else:
        assert "center_dispersion_1" in resultdata
        assert fit_data_max_at_start != fit_data_max_at_end
        if abs(fit_data_max_at_start - fit_data_max_at_end) < 3:
            warnings.warn(
                dedent(
                    """
                    Bad test, one of the following could be the case:
                    - dispersion too small
                    - spectral window to small
                    - time resolution (around the maximum of the IRF) too low"
                    """
                )
            )

        for x in suite.axis["spectral"]:
            # calculated irf location
            irf = fill_item(suite.model.irf["irf1"], suite.model, result.optimized_parameters)
            model_irf_center = irf.center
            model_dispersion_center = irf.dispersion_center
            model_center_dispersion_coefficients = irf.center_dispersion_coefficients
            calc_irf_location_at_x = _calculate_irf_position(
                x, model_irf_center, model_dispersion_center, model_center_dispersion_coefficients
            )
            # fitted irf location
            fitted_irf_loc_at_x = resultdata["irf_center_location"].sel(spectral=x)
            assert np.allclose(calc_irf_location_at_x, fitted_irf_loc_at_x.values), dedent(
                f"""
                Error in {suite.__name__} comparing irf_center_location,
                - diff={calc_irf_location_at_x-fitted_irf_loc_at_x.values}
                """
            )

    assert "irf_center" in resultdata



"""This module contains weight item."""

from glotaran.model.item import Item
from glotaran.model.item import item


@item
class Weight(Item):
    """The `Weight` class describes a value by which a dataset will scaled.

    `global_interval` and `model_interval` are optional. The whole range of the dataset
    will be used if not set.
    """

    datasets: list[str]
    global_interval: tuple[float, float] | None = None
    model_interval: tuple[float, float] | None = None
    value: float



from __future__ import annotations

import warnings
from importlib import import_module
from typing import TYPE_CHECKING

import pytest

from glotaran.deprecation.deprecation_utils import GlotaranApiDeprecationWarning
from glotaran.deprecation.deprecation_utils import module_attribute
from glotaran.optimization import optimize as optimize_module
from glotaran.simulation import simulation as simulation_module

if TYPE_CHECKING:
    from _pytest.recwarn import WarningsRecorder


def check_recwarn(records: WarningsRecorder, warn_nr=1):
    for record in records:
        print(record)

    assert len(records) == warn_nr
    assert records[0].category == GlotaranApiDeprecationWarning

    records.clear()


def changed_import_test_warn(
    recwarn: WarningsRecorder, module_qual_name: str, *, attribute_name: str = None, warn_nr=1
):
    """Helper for testing changed imports, returning the imported item.

    Parameters
    ----------
    module_qual_name : str
        Fully qualified name for a module e.g. ``glotaran.model.base_model``
    attribute_name : str, optional
        Name of the attribute e.g. ``Model``

    Returns
    -------
    Any
        Module attribute or module
    """

    warnings.simplefilter("always")

    recwarn.clear()

    if attribute_name is not None:
        result = module_attribute(module_qual_name, attribute_name)
    else:
        result = import_module(module_qual_name)
    check_recwarn(recwarn, warn_nr=warn_nr)
    return result


@pytest.mark.xfail(strict=True, reason="Fail if no warning")
def test_changed_import_test_warn_attribute_no_warn(
    recwarn: WarningsRecorder,
):
    """Module attribute import not warning"""
    changed_import_test_warn(recwarn, "glotaran.parameter", attribute_name="Parameters")


@pytest.mark.xfail(strict=True, reason="Fail if no warning")
def test_changed_import_test_warn_module_no_warn(
    recwarn: WarningsRecorder,
):
    """Module import not warning"""
    changed_import_test_warn(recwarn, "glotaran.parameter")


def test_analysis_optimization(recwarn: WarningsRecorder):
    """Usage of glotaran.analysis.optimization"""
    warnings.simplefilter("always")

    from glotaran.analysis import optimize as analysis_optimize

    assert len(recwarn) == 0
    assert analysis_optimize.optimize == optimize_module.optimize

    check_recwarn(recwarn)


def test_analysis_optimization_from_import(recwarn: WarningsRecorder):
    """Same as 'from glotaran.analysis.optimize import optimizations analysis_scheme'"""
    changed_import_test_warn(recwarn, "glotaran.analysis.optimize", attribute_name="optimize")


def test_analysis_simulation(recwarn: WarningsRecorder):
    """Usage of glotaran.analysis.simulation"""
    warnings.simplefilter("always")

    from glotaran.analysis import simulation as analysis_simulation

    assert len(recwarn) == 0
    assert analysis_simulation.simulate == simulation_module.simulate

    check_recwarn(recwarn)


def test_analysis_simulation_from_import(recwarn: WarningsRecorder):
    """Same as 'from glotaran.analysis.simulation import simulate as analysis_scheme'"""
    changed_import_test_warn(recwarn, "glotaran.analysis.simulation", attribute_name="simulate")


@pytest.mark.parametrize(
    "attribute_name", ("sim_model", "dataset", "model", "scheme", "wanted_parameter", "parameter")
)
def test_examples_sequential(recwarn: WarningsRecorder, attribute_name: str):
    """glotaran.examples.sequential exported attributes"""
    from glotaran.examples import sequential  # noqa: F401

    recwarn.clear()

    changed_import_test_warn(
        recwarn, "glotaran.examples.sequential", attribute_name=attribute_name
    )



"""Utility functions for io plugin."""

from __future__ import annotations

import os
from collections.abc import Callable
from functools import partial
from functools import wraps
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import TypeVar
from typing import cast

DecoratedFunc = TypeVar("DecoratedFunc", bound=Callable[..., Any])  # decorated function

if TYPE_CHECKING:
    from collections.abc import Iterable
    from collections.abc import Iterator

    from glotaran.typing import StrOrPath


def infer_file_format(
    file_path: StrOrPath, *, needs_to_exist: bool = True, allow_folder=False
) -> str:
    """Infer format of a file if it exists.

    Parameters
    ----------
    file_path : StrOrPath
        Path/str to the file.
    needs_to_exist : bool
        Whether or not a file need to exists for an successful format inferring.
        While write functions don't need the file to exists, load functions do.
    allow_folder: bool
        Whether or not to allow the format to be ``folder``.
        This is only used in ``save_result``.

    Returns
    -------
    str
        File extension without the leading dot.

    Raises
    ------
    ValueError
        If file doesn't exists.
    ValueError
        If file has no extension.
    """
    if not os.path.isfile(file_path) and needs_to_exist and not allow_folder:
        raise ValueError(f"There is no file {file_path!r}.")

    _, file_format = os.path.splitext(file_path)
    if file_format != "":
        file_format = file_format.lstrip(".")
        return "yaml" if file_format == "yml" else file_format

    if allow_folder:
        return "yaml"
    else:
        raise ValueError(
            f"Cannot determine format of file {file_path!r}, please provide an explicit format."
        )


def not_implemented_to_value_error(func: DecoratedFunc) -> DecoratedFunc:
    """Decorate a function to raise ValueError instead of NotImplementedError.

    This decorator is supposed to be used on functions which call functions
    that might raise a NotImplementedError, but raise ValueError instead with
    the same error text.

    Parameters
    ----------
    func : DecoratedFunc
        Function to be decorated.

    Returns
    -------
    DecoratedFunc
        Wrapped function.
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return func(*args, **kwargs)
        except NotImplementedError as error:
            raise ValueError(error.args) from error

    return cast(DecoratedFunc, wrapper)


def protect_from_overwrite(path: str | os.PathLike[str], *, allow_overwrite: bool = False) -> None:
    """Raise FileExistsError if files already exists and allow_overwrite isn't True.

    As a side effect this also creates the parent directory of a file if it does not exist.

    Parameters
    ----------
    path : str
        Path to a file or folder.
    allow_overwrite : bool
        Whether or not to allow overwriting existing files, by default False

    Raises
    ------
    FileExistsError
        If path points to an existing file.
    FileExistsError
        If path points to an existing folder which is not empty.
    """
    user_info = (
        "To protect users overwriting existing files is deactivated by default. "
        "If you are absolutely sure this is what you want and need to do you can "
        "use the argument 'allow_overwrite=True'."
    )
    path = Path(path).resolve()
    if path.parent.is_file() is False:
        path.parent.mkdir(parents=True, exist_ok=True)
    if allow_overwrite:
        return
    elif path.is_file():
        raise FileExistsError(f"The file {path!r} already exists. \n{user_info}")
    elif path.is_dir() and os.listdir(str(path)):
        raise FileExistsError(
            f"The folder {path.as_posix()!r} already exists and is not empty. \n{user_info}"
        )


def bool_str_repr(value: Any, true_repr: str = "*", false_repr: str = "/") -> Any:
    """Replace boolean value with string repr.

    This function is a helper for table representation (e.g. with tabulate)
    of boolean values.

    Parameters
    ----------
    value : Any
        Arbitrary value
    true_repr : str
        Desired repr for ``True``, by default "*"
    false_repr : str
        Desired repr for ``False``, by default "/"

    Returns
    -------
    Any
        Original value or desired repr for bool

    Examples
    --------
    >>> table_data = [["foo", True, False], ["bar", False, True]]
    >>> print(tabulate(map(lambda x: map(bool_table_repr, x), table_data)))
    ---  -  -
    foo  *  /
    bar  /  *
    ---  -  -
    """
    if value is True:
        return true_repr
    elif value is False:
        return false_repr
    else:
        return value


def bool_table_repr(
    table_data: Iterable[Iterable[Any]], true_repr: str = "*", false_repr: str = "/"
) -> Iterator[Iterator[Any]]:
    """Replace boolean value with string repr for all table values.

    This function is an implementation of :func:`bool_str_repr` for a
    2D table, for easy usage with tabulate.

    Parameters
    ----------
    table_data : Iterable[Iterable[Any]]
        Data of the table e.g. a list of lists.
    true_repr : str
        Desired repr for ``True``, by default "*"
    false_repr : str
        Desired repr for ``False``, by default "/"

    Returns
    -------
    Iterator[Iterator[Any]]
        ``table_data`` with original values or desired repr for bool

    See Also
    --------
    bool_str_repr

    Examples
    --------
    >>> table_data = [["foo", True, False], ["bar", False, True]]
    >>> print(tabulate(bool_table_repr(table_data))
    ---  -  -
    foo  *  /
    bar  /  *
    ---  -  -
    """
    bool_repr = partial(bool_str_repr, true_repr=true_repr, false_repr=false_repr)
    return ((bool_repr(value) for value in values) for values in table_data)



"""This package contains functions for model simulation and fitting."""
from glotaran.deprecation.deprecation_utils import deprecate_submodule

simulation = deprecate_submodule(
    deprecated_module_name="glotaran.analysis.simulation",
    new_module_name="glotaran.simulation.simulation",
    to_be_removed_in_version="0.8.0",
)

optimize = deprecate_submodule(
    deprecated_module_name="glotaran.analysis.optimize",
    new_module_name="glotaran.optimization.optimize",
    to_be_removed_in_version="0.8.0",
)



"""The glotaran project module."""
from __future__ import annotations

import re
from collections.abc import Mapping
from dataclasses import dataclass
from dataclasses import field
from importlib.metadata import distribution
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING
from typing import Any
from typing import Literal
from warnings import warn

import pandas as pd
import xarray as xr

from glotaran.builtin.io.yml.utils import load_dict
from glotaran.deprecation.deprecation_utils import GlotaranApiDeprecationWarning
from glotaran.deprecation.deprecation_utils import check_overdue
from glotaran.io.prepare_dataset import add_svd_to_dataset
from glotaran.model import Model
from glotaran.parameter import Parameters
from glotaran.project.project_data_registry import ProjectDataRegistry
from glotaran.project.project_model_registry import ProjectModelRegistry
from glotaran.project.project_parameter_registry import ProjectParameterRegistry
from glotaran.project.project_result_registry import ProjectResultRegistry
from glotaran.project.result import Result
from glotaran.project.scheme import Scheme
from glotaran.utils.io import make_path_absolute_if_relative
from glotaran.utils.ipython import MarkdownStr
from glotaran.utils.ipython import display_file

if TYPE_CHECKING:
    from collections.abc import Hashable

    from glotaran.typing.types import LoadableDataset

TEMPLATE = "version: {gta_version}"

PROJECT_FILE_NAME = "project.gta"


@dataclass
class Project:
    """A project represents a projectfolder on disk which contains a project file.

    A project file is a file in `yml` format with name `project.gta`
    """

    version: str = field(init=False)

    file: Path
    folder: Path = field(default=None)  # type:ignore[assignment]

    def __post_init__(self):
        """Overwrite of post init."""
        self.file = Path(self.file)
        if self.folder is None:
            self.folder = self.file.parent
        self.folder = Path(self.folder)
        self.version = distribution("pyglotaran").version

        self._data_registry = ProjectDataRegistry(self.folder)
        self._model_registry = ProjectModelRegistry(self.folder)
        self._parameter_registry = ProjectParameterRegistry(self.folder)
        self._result_registry = ProjectResultRegistry(self.folder)

    @staticmethod
    def create(folder: str | Path, allow_overwrite: bool = False) -> Project:
        """Create a new project folder and file.

        Parameters
        ----------
        folder : str | Path | None
            The folder where the project will be created. If ``None``, the current work
            directory will be used.
        allow_overwrite: bool
            Whether to overwrite an existing project file.

        Returns
        -------
        Project
            The created project.

        Raises
        ------
        FileExistsError
            Raised if the project file already exists and `allow_overwrite=False`.
        """
        project_folder = make_path_absolute_if_relative(Path(folder))
        project_folder.mkdir(parents=True, exist_ok=True)
        project_file = project_folder / PROJECT_FILE_NAME
        if project_file.exists() and not allow_overwrite:
            raise FileExistsError(
                f"Project file '{project_file}' already exist. "
                "Set `allow_overwrite=True` to overwrite."
            )
        project_file.write_text(TEMPLATE.format(gta_version=distribution("pyglotaran").version))
        return Project.open(project_file, create_if_not_exist=True)

    @classmethod
    def open(cls, project_folder_or_file: str | Path, create_if_not_exist: bool = True) -> Project:
        """Open a new project.

        Parameters
        ----------
        project_folder_or_file : str | Path
            The path to a project folder or file.
        create_if_not_exist : bool
            Create the project if not existent.

        Returns
        -------
        Project
            The project instance.

        Raises
        ------
        FileNotFoundError
            Raised when the project file does not not exist and `create_if_not_exist` is `False`.
        """
        folder = make_path_absolute_if_relative(Path(project_folder_or_file))
        if folder.name == PROJECT_FILE_NAME:
            folder, project_file = folder.parent, folder
        else:
            project_file = folder / PROJECT_FILE_NAME

        if project_file.is_file() is False:
            if create_if_not_exist is False:
                raise FileNotFoundError(f"Project file {project_file.as_posix()} does not exists.")
            Project.create(folder)

        project_dict = load_dict(project_file, True)
        project_dict["file"] = project_file
        project_dict["folder"] = folder
        version = project_dict.pop("version")
        project = cls(**project_dict)
        project.version = version
        return project

    @property
    def has_data(self) -> bool:
        """Check if the project has datasets.

        Returns
        -------
        bool
            Whether the project has datasets.
        """
        return not self._data_registry.empty

    @property
    def data(self) -> Mapping[str, Path]:
        """Get all project datasets.

        Returns
        -------
        Mapping[str, Path]
            The models of the datasets.
        """
        return self._data_registry.items

    def load_data(
        self,
        dataset_name: str,
        *,
        add_svd: bool = False,
        lsv_dim: Hashable = "time",
        rsv_dim: Hashable = "spectral",
    ) -> xr.Dataset:
        """Load a dataset, with SVD data if ``add_svd`` is ``True``.

        Parameters
        ----------
        dataset_name : str
            The name of the dataset.
        add_svd: bool
            Whether or not to calculate and add SVD data. Defaults to False.
        lsv_dim: Hashable
            Dimension of the left singular vectors. Defaults to "time".
        rsv_dim: Hashable
            Dimension of the right singular vectors. Defaults to "spectral",

        Returns
        -------
        xr.Dataset
            The loaded dataset, with SVD data if ``add_svd`` is ``True``.

        Raises
        ------
        ValueError
            Raised if the dataset does not exist.


        .. # noqa: DAR402
        """
        dataset = self._data_registry.load_item(dataset_name)
        if isinstance(dataset, xr.DataArray):
            dataset = dataset.to_dataset(name="data")
        if add_svd is True:
            add_svd_to_dataset(dataset, name="data", lsv_dim=lsv_dim, rsv_dim=rsv_dim)
        return dataset

    def import_data(
        self,
        dataset: LoadableDataset | Mapping[str, LoadableDataset],
        dataset_name: str | None = None,
        allow_overwrite: bool = False,
        ignore_existing: bool = True,
    ):
        """Import a dataset by saving it as an .nc file in the project's data folder.

        Parameters
        ----------
        dataset : LoadableDataset
            Dataset instance or path to a dataset.
        dataset_name : str | None
            The name of the dataset (needs to be provided when dataset is an xarray instance).
            Defaults to None.
        allow_overwrite: bool
            Whether to overwrite an existing dataset.
        ignore_existing: bool
            Whether to skip import if the dataset already exists and allow_overwrite is False.
            Defaults to ``True``.
        """
        if not isinstance(dataset, Mapping) or isinstance(dataset, (xr.Dataset, xr.DataArray)):
            dataset = {dataset_name: dataset}

        for key, value in dataset.items():
            self._data_registry.import_data(
                value,
                dataset_name=key,
                allow_overwrite=allow_overwrite,
                ignore_existing=ignore_existing,
            )

    @property
    def has_models(self) -> bool:
        """Check if the project has models.

        Returns
        -------
        bool
            Whether the project has models.
        """
        return not self._model_registry.empty

    @property
    def models(self) -> Mapping[str, Path]:
        """Get all project models.

        Returns
        -------
        Mapping[str, Path]
            The models of the project.
        """
        return self._model_registry.items

    def load_model(self, model_name: str) -> Model:
        """Load a model.

        Parameters
        ----------
        model_name : str
            The name of the model.

        Returns
        -------
        Model
            The loaded model.

        Raises
        ------
        ValueError
            Raised if the model does not exist.


        .. # noqa: DAR402
        """
        return self._model_registry.load_item(model_name)

    def show_model_definition(self, model_name: str, syntax: str | None = None) -> MarkdownStr:
        """Show model definition file content with syntax highlighting.

        Parameters
        ----------
        model_name: str
            The name of the model.
        syntax: str | None
            Syntax used for syntax highlighting. Defaults to None which means that the syntax is
            inferred based on the file extension. Pass the value ``""`` to deactivate syntax
            highlighting.

        Returns
        -------
        MarkdownStr
            Model definition file content with syntax highlighting to render in ipython.
        """
        return display_file(self.models[model_name], syntax=syntax)

    def validate(self, model_name: str, parameters_name: str | None = None) -> MarkdownStr:
        """Check that the model is valid, list all issues in the model if there are any.

        If ``parameters_name`` also consider the ``Parameters`` when validating.

        Parameters
        ----------
        model_name: str
            The name of the model to validate.
        parameters_name: str | None
            The name of the parameters to use when validating. Defaults to ``None`` which means
            that parameters are not considered when validating the model.

        Returns
        -------
        MarkdownStr
            Text indicating if the model is valid or not.
        """
        model = self.load_model(model_name)
        return model.validate(
            self.load_parameters(parameters_name) if parameters_name is not None else None
        )

    def generate_model(
        self,
        model_name: str,
        generator_name: str,
        generator_arguments: dict[str, Any],
        *,
        allow_overwrite: bool = False,
        ignore_existing: bool = False,
    ):
        """Generate a model.

        Parameters
        ----------
        model_name : str
            The name of the model.
        generator_name : str
            The generator for the model.
        generator_arguments : dict[str, Any]
            Arguments for the generator.
        allow_overwrite: bool
            Whether to overwrite an existing model.
        ignore_existing: bool
            Whether to ignore generation of a model file if it already exists.
        """
        check_overdue("Project.generate_model", "0.8.0")
        warn(
            GlotaranApiDeprecationWarning(
                "Usage of 'Project.generate_model' was deprecated without replacement.\n"
                "This usage will be an error in version: '0.8.0'."
            ),
            stacklevel=2,
        )
        self._model_registry.generate_model(
            model_name,
            generator_name,
            generator_arguments,  # type:ignore[arg-type]
            allow_overwrite=allow_overwrite,
            ignore_existing=ignore_existing,
        )

    def get_models_directory(self) -> Path:
        """Get the path to the model directory of the project.

        Returns
        -------
        Path
            The path to the project's model directory.
        """
        return self._model_registry.directory

    @property
    def has_parameters(self) -> bool:
        """Check if the project has parameters.

        Returns
        -------
        bool
            Whether the project has parameters.
        """
        return not self._parameter_registry.empty

    @property
    def parameters(self) -> Mapping[str, Path]:
        """Get all project parameters.

        Returns
        -------
        Mapping[str, Path]
            The parameters of the project.
        """
        return self._parameter_registry.items

    def load_parameters(self, parameters_name: str) -> Parameters:
        """Load parameters.

        Parameters
        ----------
        parameters_name : str
            The name of the parameters.

        Raises
        ------
        ValueError
            Raised if parameters do not exist.

        Returns
        -------
        Parameters
            The loaded parameters.


        .. # noqa: D414
        .. # noqa: DAR402
        """
        return self._parameter_registry.load_item(parameters_name)

    def show_parameters_definition(
        self, parameters_name: str, syntax: str | None = None, *, as_dataframe: bool | None = None
    ) -> MarkdownStr | pd.DataFrame:
        """Show parameters definition file content with syntax highlighting.

        Parameters
        ----------
        parameters_name: str
            The name of the parameters.
        syntax: str | None
            Syntax used for syntax highlighting. Defaults to None which means that the syntax is
            inferred based on the file extension. Pass the value ``""`` to deactivate syntax
            highlighting.
        as_dataframe: bool | None
            Whether or not to show the ``Parameters`` definition as pandas.DataFrame (mostly useful
            for non string formats). Defaults to None which means that it will be inferred to
            ``True`` for known non string formats like ``xlsx``.

        Returns
        -------
        MarkdownStr | pd.DataFrame
            Parameters definition file content with syntax highlighting to render in ipython.
        """
        if as_dataframe is True or (
            as_dataframe is None and self.parameters[parameters_name].suffix in [".xlsx", ".ods"]
        ):
            return self.load_parameters(parameters_name).to_dataframe()
        return display_file(self.parameters[parameters_name], syntax=syntax)

    def generate_parameters(
        self,
        model_name: str,
        parameters_name: str | None = None,
        *,
        format_name: Literal["yml", "yaml", "csv"] = "csv",
        allow_overwrite: bool = False,
        ignore_existing: bool = False,
    ):
        """Generate parameters for a model.

        Parameters
        ----------
        model_name : str
            The model.
        parameters_name : str | None
            The name of the parameters. If ``None`` it will be <model_name>_parameters.
        format_name : Literal["yml", "yaml", "csv"]
            The parameter format.
        allow_overwrite: bool
            Whether to overwrite existing parameters.
        ignore_existing: bool
            Whether to ignore generation of a parameter file if it already exists.
        """
        check_overdue("Project.generate_parameters", "0.8.0")
        warn(
            GlotaranApiDeprecationWarning(
                "Usage of 'Project.generate_parameters' was deprecated without replacement.\n"
                "This usage will be an error in version: '0.8.0'."
            ),
            stacklevel=2,
        )
        model = self.load_model(model_name)
        parameters_name = (
            parameters_name if parameters_name is not None else f"{model_name}_parameters"
        )
        self._parameter_registry.generate_parameters(
            model,
            parameters_name,
            format_name=format_name,
            allow_overwrite=allow_overwrite,
            ignore_existing=ignore_existing,
        )

    def get_parameters_directory(self) -> Path:
        """Get the path to the parameter directory of the project.

        Returns
        -------
        Path
            The path to the project's parameter directory.
        """
        return self._parameter_registry.directory

    @property
    def has_results(self) -> bool:
        """Check if the project has results.

        Returns
        -------
        bool
            Whether the project has results.
        """
        return not self._result_registry.empty

    @property
    def results(self) -> Mapping[str, Path]:
        """Get all project results.

        Returns
        -------
        Mapping[str, Path]
            The results of the project.
        """
        return self._result_registry.items

    def get_result_path(self, result_name: str, *, latest: bool = False) -> Path:
        """Get the path to a result with name ``name``.

        Parameters
        ----------
        result_name : str
            The name of the result.
        latest: bool
            Flag to deactivate warning about using latest result. Defaults to False

        Returns
        -------
        Path
            The path to the result.

        Raises
        ------
        ValueError
            Raised if result does not exist.


        .. # noqa: DAR402
        """
        return self._result_registry._latest_result_path_fallback(result_name, latest=latest)

    def get_latest_result_path(self, result_name: str) -> Path:
        """Get the path to a result with name ``name``.

        Parameters
        ----------
        result_name : str
            The name of the result.

        Returns
        -------
        Path
            The path to the result.

        Raises
        ------
        ValueError
            Raised if result does not exist.


        .. # noqa: DAR402
        """
        result_name = re.sub(self._result_registry.result_pattern, "", result_name)
        return self.get_result_path(result_name, latest=True)

    def load_result(self, result_name: str, *, latest: bool = False) -> Result:
        """Load a result.

        Parameters
        ----------
        result_name : str
            The name of the result.
        latest: bool
            Flag to deactivate warning about using latest result. Defaults to False

        Returns
        -------
        Result
            The loaded result.

        Raises
        ------
        ValueError
            Raised if result does not exist.


        .. # noqa: DAR402
        """
        return self._result_registry._loader(
            self._result_registry._latest_result_path_fallback(result_name, latest=latest)
        )

    def load_latest_result(self, result_name: str) -> Result:
        """Load a result.

        Parameters
        ----------
        result_name : str
            The name of the result.

        Returns
        -------
        Result
            The loaded result.

        Raises
        ------
        ValueError
            Raised if result does not exist.


        .. # noqa: DAR402
        """
        result_name = re.sub(self._result_registry.result_pattern, "", result_name)
        return self.load_result(result_name, latest=True)

    def create_scheme(
        self,
        model_name: str,
        parameters_name: str,
        maximum_number_function_evaluations: int | None = None,
        clp_link_tolerance: float = 0.0,
        data_lookup_override: Mapping[str, LoadableDataset] | None = None,
    ) -> Scheme:
        """Create a scheme for optimization.

        Parameters
        ----------
        model_name : str
            The model to optimize.
        parameters_name : str
            The initial parameters.
        maximum_number_function_evaluations : int | None
            The maximum number of function evaluations.
        clp_link_tolerance : float
            The CLP link tolerance.
        data_lookup_override: Mapping[str, LoadableDataset] | None
            Allows to bypass the default dataset lookup in the project ``data`` folder and use a
            different dataset for the optimization without changing the model. This is especially
            useful when working with preprocessed data. Defaults to ``None``.

        Returns
        -------
        Scheme
            The created scheme.
        """
        if data_lookup_override is None:
            data_lookup_override = {}
        loaded_model = self.load_model(model_name)
        data_lookup_override = {
            dataset_name: dataset_value
            for dataset_name, dataset_value in data_lookup_override.items()
            if dataset_name in loaded_model.dataset
        }
        data = {
            dataset_name: self.data[dataset_name]
            for dataset_name in loaded_model.dataset
            if dataset_name not in data_lookup_override
        }
        return Scheme(
            model=loaded_model,
            parameters=self.load_parameters(parameters_name),
            data=data | data_lookup_override,
            maximum_number_function_evaluations=maximum_number_function_evaluations,
            clp_link_tolerance=clp_link_tolerance,
        )

    def optimize(
        self,
        model_name: str,
        parameters_name: str,
        result_name: str | None = None,
        maximum_number_function_evaluations: int | None = None,
        clp_link_tolerance: float = 0.0,
        data_lookup_override: Mapping[str, LoadableDataset] | None = None,
    ) -> Result:
        """Optimize a model.

        Parameters
        ----------
        model_name : str
            The model to optimize.
        parameters_name : str
            The initial parameters.
        result_name : str | None
            The name of the result.
        maximum_number_function_evaluations : int | None
            The maximum number of function evaluations.
        clp_link_tolerance : float
            The CLP link tolerance.
        data_lookup_override: Mapping[str, LoadableDataset] | None
            Allows to bypass the default dataset lookup in the project ``data`` folder and use a
            different dataset for the optimization without changing the model. This is especially
            useful when working with preprocessed data. Defaults to ``None``.

        Returns
        -------
        Result
            Result of the optimization.
        """
        from glotaran.optimization.optimize import optimize

        scheme = self.create_scheme(
            model_name=model_name,
            parameters_name=parameters_name,
            maximum_number_function_evaluations=maximum_number_function_evaluations,
            clp_link_tolerance=clp_link_tolerance,
            data_lookup_override=data_lookup_override,
        )
        result = optimize(scheme)

        result_name = result_name or model_name
        self._result_registry.save(result_name, result)
        return result

    def markdown(self) -> MarkdownStr:
        """Format the project as a markdown text.

        Returns
        -------
        MarkdownStr : str
            The markdown string.
        """
        md = f"""\
            # Project (_{self.folder.name}_)

            pyglotaran version: `{self.version}`

            ## Data

            {self._data_registry.markdown(join_indentation=12)}

            ## Model

            {self._model_registry.markdown(join_indentation=12)}

            ## Parameters

            {self._parameter_registry.markdown(join_indentation=12)}

            ## Results

            {self._result_registry.markdown(join_indentation=12)}
            """

        return MarkdownStr(dedent(md))

    def _repr_markdown_(self) -> str:
        """Create a markdown representation.

        Special method used by ``ipython`` to render markdown.

        Returns
        -------
        str :
            The markdown representation as string.
        """
        return str(self.markdown())



from IPython.core.formatters import format_display_data

from glotaran.io import load_parameters
from glotaran.parameter.parameters import Parameters

PARAMETERS_3C_BASE = """\
irf:
    - ["center", 1.3]
    - ["width", 7.8]
j:
    - ["1", 1, {"vary": False, "non-negative": False}]
"""

PARAMETERS_3C_KINETIC = """\
kinetic:
    - ["1", 300e-3]
    - ["2", 500e-4, {standard-error: 0.000012345678}]
    - ["3", {expr: $kinetic.1 + $kinetic.2}]
"""

RENDERED_MARKDOWN = """\
  * __irf__:

    | _Label_   |   _Value_ |   _Standard Error_ | _t-value_   |   _Minimum_ |   _Maximum_ | _Vary_   | _Non-Negative_   | _Expression_   |
    |-----------|-----------|--------------------|-------------|-------------|-------------|----------|------------------|----------------|
    | center    | 1.300e+00 |                nan |  nan        |        -inf |         inf | True     | False            | `None`         |
    | width     | 7.800e+00 |                nan |  nan        |        -inf |         inf | True     | False            | `None`         |

  * __j__:

    |   _Label_ |   _Value_ |   _Standard Error_ | _t-value_   |   _Minimum_ |   _Maximum_ | _Vary_   | _Non-Negative_   | _Expression_   |
    |-----------|-----------|--------------------|-------------|-------------|-------------|----------|------------------|----------------|
    |         1 | 1.000e+00 |                nan |  nan        |        -inf |         inf | False    | False            | `None`         |

  * __kinetic__:

    |   _Label_ |   _Value_ |   _Standard Error_ | _t-value_   |   _Minimum_ |   _Maximum_ | _Vary_   | _Non-Negative_   | _Expression_              |
    |-----------|-----------|--------------------|-------------|-------------|-------------|----------|------------------|---------------------------|
    |         1 | 3.000e-01 |        nan         |  nan        |        -inf |         inf | True     | False            | `None`                    |
    |         2 | 5.000e-02 |          1.235e-05 |  4050       |        -inf |         inf | True     | False            | `None`                    |
    |         3 | 3.500e-01 |        nan         |  nan        |        -inf |         inf | False    | False            | `$kinetic.1 + $kinetic.2` |

"""  # noqa: E501

RENDERED_MARKDOWN_E5_PRECISION = """\
  * __irf__:

    | _Label_   |     _Value_ |   _Standard Error_ | _t-value_   |   _Minimum_ |   _Maximum_ | _Vary_   | _Non-Negative_   | _Expression_   |
    |-----------|-------------|--------------------|-------------|-------------|-------------|----------|------------------|----------------|
    | center    | 1.30000e+00 |        1.23457e-05 |  105300     |        -inf |         inf | True     | False            | `None`         |

"""  # noqa: E501


def test_parameters_markdown_is_order_independent():
    """Markdown output of Parameters.markdown() is independent of initial order"""
    PARAMETERS_3C_INITIAL1 = f"""{PARAMETERS_3C_BASE}\n{PARAMETERS_3C_KINETIC}"""
    PARAMETERS_3C_INITIAL2 = f"""{PARAMETERS_3C_KINETIC}\n{PARAMETERS_3C_BASE}"""

    initial_parameters_ref = Parameters.from_dict(
        {
            "j": [["1", 1, {"vary": False, "non-negative": False}]],
            "kinetic": [
                ["1", 0.3],
                ["2", 500e-4, {"standard-error": 0.000012345678}],
                ["3", 700e-5, {"expr": "$kinetic.1 + $kinetic.2"}],
            ],
            "irf": [["center", 1.3], ["width", 7.8]],
        }
    )

    initial_parameters1 = load_parameters(PARAMETERS_3C_INITIAL1, format_name="yml_str")
    initial_parameters2 = load_parameters(PARAMETERS_3C_INITIAL2, format_name="yml_str")

    assert str(initial_parameters1.markdown()) == RENDERED_MARKDOWN
    assert str(initial_parameters2.markdown()) == RENDERED_MARKDOWN
    assert str(initial_parameters_ref.markdown()) == RENDERED_MARKDOWN

    minimal_params = Parameters.from_dict(
        {"irf": [["center", 1.3, {"standard-error": 0.000012345678}]]}
    )

    assert str(minimal_params.markdown(float_format=".5e")) == RENDERED_MARKDOWN_E5_PRECISION


def test_parameters_repr():
    """Repr creates code to recreate the object with from_dict."""

    # Needed to eval the Parameters repr
    from glotaran.parameter.parameter import Parameter  # noqa:401

    result = Parameters.from_dict(
        {
            "foo": {
                "bar": [
                    ["1", 1.0, {"vary": True}],
                    ["2", 2.0, {"expression": "$foo.bar.1*2"}],
                    ["3", 3.0, {"min": -10}],
                ]
            }
        }
    )
    expected = (
        "Parameters({'foo.bar.1': Parameter(label='foo.bar.1', value=1.0), "
        "'foo.bar.2': Parameter(label='foo.bar.2', value=2.0, expression='$foo.bar.1*2',"
        " vary=False), "
        "'foo.bar.3': Parameter(label='foo.bar.3', value=3.0, minimum=-10)})"
    )

    print(result.__repr__())
    assert result.__repr__() == expected
    assert result == eval(expected)


def test_parameters_ipython_rendering():
    """Autorendering in ipython"""
    param_group = Parameters.from_dict({"foo": {"bar": [["1", 1.0], ["2", 2.0], ["3", 3.0]]}})
    rendered_obj = format_display_data(param_group)[0]

    assert "text/markdown" in rendered_obj
    assert rendered_obj["text/markdown"].startswith("  * __foo__")

    rendered_markdown_return = format_display_data(param_group.markdown())[0]

    assert "text/markdown" in rendered_markdown_return
    assert rendered_markdown_return["text/markdown"].startswith("  * __foo__")



import click
from click import echo
from click import pause
from click import prompt

from glotaran.cli.commands import util


@click.option(
    "--name",
    "-n",
    default=None,
    type=str,
    show_default=True,
    help="Name of the datagroup to export.",
)
@click.option(
    "--select",
    "-s",
    default=None,
    type=(str, util.VALORRANGEORLIST),
    show_default=True,
    multiple=True,
    help="Selection of data. Example --select DIMENSION_NAME VALUE "
    "where VALUE can be a scalar, a list like '[V1,V2,V3]' or a tuple like '(MIN,MAX)'",
)
@click.option(
    "--out",
    "-o",
    default=None,
    type=click.Path(dir_okay=False),
    show_default=True,
    help="Path to the output file.",
)
@click.argument("filename", type=click.Path(exists=True))
def export(filename: str, select, out: str, name: str):
    """Exports data from netCDF4 to ascii."""

    echo(f"Opening dataset at {filename}")

    dataset = util.load_dataset(filename, dtype="nc")

    if name is None:
        name = util.select_name(filename, dataset)

    stop = False
    data = dataset[name]

    for sel in select:
        data = util.select_data(data, sel[0], sel[1])

    if out is not None:
        if len(data.shape) > 2:
            pause("Cannot export data with more than 2 dimensions to ASCII")
        else:
            util.write_data(data, out)
            stop = True

    while not stop:
        echo(f"Selected dataset '{name}'.")
        echo(f"\nDataset Content\n\n{data}\n")

        choice = prompt(
            "How to proceed",
            default="exit",
            type=click.Choice(["exit", "save", "select", "change", "help"]),
        )
        if choice == "exit":
            stop = True
        elif choice == "change":
            name = util.select_name(filename, dataset)
            data = dataset[name]
        elif choice == "select":
            cont = True
            while cont:
                dims = list(data.dims)
                choice = prompt(
                    "Please select a dimension or action",
                    default="back",
                    type=click.Choice(dims + ["reset", "back"]),
                )
                if choice == "back":
                    cont = False
                elif choice == "reset":
                    data = dataset[name]
                else:
                    dim = choice
                    choice = prompt(
                        "Please select a value. Type 2 values sperated by ',' to select a range.",
                        default="back",
                        type=util.VALORRANGEORLIST,
                    )
                    if choice != "back":
                        try:
                            data = util.select_data(data, dim, choice)
                        except ValueError as e:
                            echo(e, err=True)
                            continue

        elif choice == "save":
            if len(data.shape) > 2:
                pause("Cannot export data with more than 2 dimensions to ASCII")
            else:
                cont = True
                change = False
                while cont:
                    if out is None or change:
                        out = prompt(
                            "Please type in a name for the output file.",
                            default="out.csv",
                            type=click.Path(dir_okay=False),
                        )
                        change = False
                    choice = prompt(
                        f"The filepath of the output file is '{out}'. Save the data?",
                        default="yes",
                        type=click.Choice(["yes", "no", "change"]),
                    )
                    if choice == "change":
                        change = True
                    else:
                        cont = False
                        if choice == "yes":
                            util.write_data(data, out)
    echo("Good-bye, have a nice day!")



"""This module contains the dataset model."""

from __future__ import annotations

from collections.abc import Generator
from typing import TYPE_CHECKING

import xarray as xr

from glotaran.model.item import ItemIssue
from glotaran.model.item import ModelItem
from glotaran.model.item import ModelItemType
from glotaran.model.item import ParameterType
from glotaran.model.item import attribute
from glotaran.model.item import item
from glotaran.model.megacomplex import Megacomplex
from glotaran.model.megacomplex import is_exclusive
from glotaran.model.megacomplex import is_unique

if TYPE_CHECKING:
    from glotaran.model.model import Model
    from glotaran.parameter import Parameter
    from glotaran.parameter import Parameters


class ExclusiveMegacomplexIssue(ItemIssue):
    """Issue for exclusive megacomplexes."""

    def __init__(self, label: str, megacomplex_type: str, is_global: bool):
        """Create an ExclusiveMegacomplexIssue.

        Parameters
        ----------
        label : str
            The megacomplex label.
        megacomplex_type : str
            The megacomplex type.
        is_global : bool
            Whether the megacomplex is global.
        """
        self._label = label
        self._type = megacomplex_type
        self._is_global = is_global

    def to_string(self) -> str:
        """Get the issue as string.

        Returns
        -------
        str
        """
        return (
            f"Exclusive {'global ' if self._is_global else ''}megacomplex '{self._label}' of "
            f"type '{self._type}' cannot be combined with other megacomplexes."
        )


class UniqueMegacomplexIssue(ItemIssue):
    """Issue for unique megacomplexes."""

    def __init__(self, label: str, megacomplex_type: str, is_global: bool):
        """Create a UniqueMegacomplexIssue.

        Parameters
        ----------
        label : str
            The megacomplex label.
        megacomplex_type : str
            The megacomplex type.
        is_global : bool
            Whether the megacomplex is global.
        """
        self._label = label
        self._type = megacomplex_type
        self._is_global = is_global

    def to_string(self):
        """Get the issue as string.

        Returns
        -------
        str
        """
        return (
            f"Unique {'global ' if self._is_global else ''}megacomplex '{self._label}' of "
            f"type '{self._type}' can only be used once per dataset."
        )


def get_megacomplex_issues(
    value: list[str | Megacomplex] | None, model: Model, is_global: bool
) -> list[ItemIssue]:
    """Get issues for megacomplexes.

    Parameters
    ----------
    value: list[str | Megacomplex] | None
        A list of megacomplexes.
    model: Model
        The model.
    is_global: bool
        Whether the megacomplexes are global.

    Returns
    -------
    list[ItemIssue]
    """
    issues: list[ItemIssue] = []

    if value is not None:
        labels = [v if isinstance(v, str) else v.label for v in value]
        megacomplexes = [model.megacomplex[label] for label in labels]
        for megacomplex in megacomplexes:
            megacomplex_type = megacomplex.__class__
            if is_exclusive(megacomplex_type) and len(megacomplexes) > 1:
                issues.append(
                    ExclusiveMegacomplexIssue(megacomplex.label, megacomplex.type, is_global)
                )
            if (
                is_unique(megacomplex_type)
                and len([m for m in megacomplexes if m.__class__ is megacomplex_type]) > 1
            ):
                issues.append(
                    UniqueMegacomplexIssue(megacomplex.label, megacomplex.type, is_global)
                )
    return issues


def validate_megacomplexes(
    value: list[str | Megacomplex],
    dataset_model: DatasetModel,
    model: Model,
    parameters: Parameters | None,
) -> list[ItemIssue]:
    """Get issues for dataset model megacomplexes.

    Parameters
    ----------
    value: list[str | Megacomplex]
        A list of megacomplexes.
    dataset_model: DatasetModel
        The dataset model.
    model: Model
        The model.
    parameters: Parameters | None,
        The parameters.

    Returns
    -------
    list[ItemIssue]
    """
    return get_megacomplex_issues(value, model, False)


def validate_global_megacomplexes(
    value: list[str | Megacomplex] | None,
    dataset_model: DatasetModel,
    model: Model,
    parameters: Parameters | None,
) -> list[ItemIssue]:
    """Get issues for dataset model global megacomplexes.

    Parameters
    ----------
    value: list[str | Megacomplex] | None
        A list of megacomplexes.
    dataset_model: DatasetModel
        The dataset model.
    model: Model
        The model.
    parameters: Parameters | None,
        The parameters.

    Returns
    -------
    list[ItemIssue]
    """
    return get_megacomplex_issues(value, model, False)


@item
class DatasetModel(ModelItem):
    """A model for datasets."""

    group: str = "default"
    force_index_dependent: bool = False
    megacomplex: list[ModelItemType[Megacomplex]] = attribute(
        validator=validate_megacomplexes  # type:ignore[arg-type]
    )
    megacomplex_scale: list[ParameterType] | None = None
    global_megacomplex: list[ModelItemType[Megacomplex]] | None = attribute(
        alias="megacomplex",
        default=None,
        validator=validate_global_megacomplexes,  # type:ignore[arg-type]
    )
    global_megacomplex_scale: list[ParameterType] | None = None
    scale: ParameterType | None = None


def has_dataset_model_global_model(dataset_model: DatasetModel) -> bool:
    """Check if the dataset model can model the global dimension.

    Parameters
    ----------
    dataset_model: DatasetModel
        The dataset model.

    Returns
    -------
    bool
    """
    return (
        dataset_model.global_megacomplex is not None and len(dataset_model.global_megacomplex) != 0
    )


def get_dataset_model_model_dimension(dataset_model: DatasetModel) -> str:
    """Get the dataset model's model dimension.

    Parameters
    ----------
    dataset_model: DatasetModel
        The dataset model.

    Returns
    -------
    str

    Raises
    ------
    ValueError
        Raised if the dataset model does not have megacomplexes or if it is not filled.
    """
    if len(dataset_model.megacomplex) == 0:
        raise ValueError(f"No megacomplex set for dataset model '{dataset_model.label}'.")
    if any(isinstance(m, str) for m in dataset_model.megacomplex):
        raise ValueError(f"Dataset model '{dataset_model.label}' was not filled.")
    model_dimension: str = dataset_model.megacomplex[
        0
    ].dimension  # type:ignore[union-attr, assignment]
    if any(
        model_dimension != m.dimension  # type:ignore[union-attr]
        for m in dataset_model.megacomplex
    ):
        raise ValueError(
            f"Megacomplex dimensions do not match for dataset model '{dataset_model.label}'."
        )
    return model_dimension


def iterate_dataset_model_megacomplexes(
    dataset_model: DatasetModel,
) -> Generator[tuple[Parameter | str | None, Megacomplex | str], None, None]:
    """Iterate the dataset model's megacomplexes.

    Parameters
    ----------
    dataset_model: DatasetModel
        The dataset model.

    Yields
    ------
    tuple[Parameter | str | None, Megacomplex | str]
        A scale and megacomplex.
    """
    for i, megacomplex in enumerate(dataset_model.megacomplex):
        scale = (
            dataset_model.megacomplex_scale[i]
            if dataset_model.megacomplex_scale is not None
            else None
        )
        yield scale, megacomplex


def iterate_dataset_model_global_megacomplexes(
    dataset_model: DatasetModel,
) -> Generator[tuple[Parameter | str | None, Megacomplex | str], None, None]:
    """Iterate the dataset model's global megacomplexes.

    Parameters
    ----------
    dataset_model: DatasetModel
        The dataset model.

    Yields
    ------
    tuple[Parameter | str | None, Megacomplex | str]
        A scale and megacomplex.
    """
    if dataset_model.global_megacomplex is None:
        return
    for i, megacomplex in enumerate(dataset_model.global_megacomplex):
        scale = (
            dataset_model.global_megacomplex_scale[i]
            if dataset_model.global_megacomplex_scale is not None
            else None
        )
        yield scale, megacomplex


def finalize_dataset_model(dataset_model: DatasetModel, dataset: xr.Dataset):
    """Finalize a dataset by applying all megacomplex finalize methods.

    Parameters
    ----------
    dataset_model: DatasetModel
        The dataset model.
    dataset: xr.Dataset
        The dataset.
    """
    is_full_model = has_dataset_model_global_model(dataset_model)
    for megacomplex in dataset_model.megacomplex:
        megacomplex.finalize_data(  # type:ignore[union-attr]
            dataset_model, dataset, is_full_model=is_full_model
        )
    if is_full_model and dataset_model.global_megacomplex is not None:
        for megacomplex in dataset_model.global_megacomplex:
            megacomplex.finalize_data(  # type:ignore[union-attr]
                dataset_model, dataset, is_full_model=is_full_model, as_global=True
            )



from __future__ import annotations

from pathlib import Path
from textwrap import dedent

import pytest

from glotaran.builtin.megacomplexes.decay import DecayMegacomplex
from glotaran.model import Megacomplex
from glotaran.model import megacomplex
from glotaran.plugin_system.base_registry import PluginOverwriteWarning
from glotaran.plugin_system.base_registry import __PluginRegistry
from glotaran.plugin_system.megacomplex_registration import get_megacomplex
from glotaran.plugin_system.megacomplex_registration import is_known_megacomplex
from glotaran.plugin_system.megacomplex_registration import known_megacomplex_names
from glotaran.plugin_system.megacomplex_registration import megacomplex_plugin_table
from glotaran.plugin_system.megacomplex_registration import register_megacomplex
from glotaran.plugin_system.megacomplex_registration import set_megacomplex_plugin
from glotaran.testing.plugin_system import monkeypatch_plugin_registry_megacomplex

MOCK_REGISTRY_VALUES = {
    "foo": Megacomplex,
    "bar": DecayMegacomplex,
    "glotaran.builtin.megacomplexes.decay.DecayMegacomplex": DecayMegacomplex,
}


@pytest.fixture
def mocked_registry():
    with monkeypatch_plugin_registry_megacomplex(MOCK_REGISTRY_VALUES, create_new_registry=True):
        yield


@pytest.mark.usefixtures("mocked_registry")
def test_register_megacomplex():
    """Register new megacomplex."""
    register_megacomplex("base-megacomplex", Megacomplex)

    assert "base-megacomplex" in __PluginRegistry.megacomplex
    assert __PluginRegistry.megacomplex["base-megacomplex"] == Megacomplex
    assert "glotaran.model.megacomplex.Megacomplex" in __PluginRegistry.megacomplex
    assert __PluginRegistry.megacomplex["glotaran.model.megacomplex.Megacomplex"] == Megacomplex
    assert known_megacomplex_names(full_names=True) == sorted(
        [
            "foo",
            "bar",
            "glotaran.builtin.megacomplexes.decay.DecayMegacomplex",
            "base-megacomplex",
            "glotaran.model.megacomplex.Megacomplex",
        ]
    )


@pytest.mark.usefixtures("mocked_registry")
def test_register_megacomplex_warning():
    """PluginOverwriteWarning raised pointing to correct file."""

    with pytest.warns(PluginOverwriteWarning, match="DecayMegacomplex.+bar.+Dummy") as record:

        @megacomplex()
        class Dummy(DecayMegacomplex):
            type: str = "bar"

        assert len(record) == 1
        assert Path(record[0].filename) == Path(__file__)


@pytest.mark.usefixtures("mocked_registry")
def test_is_known_megacomplex():
    """Check if megacomplexs are in registry"""
    assert is_known_megacomplex("foo")
    assert is_known_megacomplex("bar")
    assert not is_known_megacomplex("baz")


@pytest.mark.usefixtures("mocked_registry")
def test_get_megacomplex():
    """Get megacomplex from registry"""
    assert get_megacomplex("foo") == Megacomplex


@pytest.mark.usefixtures("mocked_registry")
def test_known_megacomplex_names():
    """Get megacomplex names from registry"""
    assert known_megacomplex_names() == sorted(["foo", "bar"])


@pytest.mark.usefixtures("mocked_registry")
def test_known_set_megacomplex_plugin():
    """Overwrite foo megacomplex"""
    assert get_megacomplex("foo") == Megacomplex
    set_megacomplex_plugin("foo", "glotaran.builtin.megacomplexes.decay.DecayMegacomplex")
    assert get_megacomplex("foo") == DecayMegacomplex


@pytest.mark.usefixtures("mocked_registry")
def test_known_set_megacomplex_plugin_dot_in_megacomplex_name():
    """Raise error if megacomplex_name contains '.'"""
    with pytest.raises(
        ValueError,
        match=r"The value of 'megacomplex_name' isn't allowed to contain the character '\.' \.",
    ):
        set_megacomplex_plugin("foo.bar", "glotaran.builtin.megacomplexes.decay.DecayMegacomplex")


@pytest.mark.usefixtures("mocked_registry")
def test_megacomplex_plugin_table():
    """Short megacomplex table."""
    expected = dedent(
        """\
        |  __Megacomplex name__  |
        |------------------------|
        |         `bar`          |
        |         `foo`          |
        """
    )
    print(f"{megacomplex_plugin_table()}\n")
    assert f"{megacomplex_plugin_table()}\n" == expected


@pytest.mark.usefixtures("mocked_registry")
def test_megacomplex_plugin_table_full():
    """Full Table with all extras."""
    expected = dedent(
        """\
        |                  __Megacomplex name__                   |                              __Plugin name__                              |
        |---------------------------------------------------------|---------------------------------------------------------------------------|
        |                          `bar`                          | `glotaran.builtin.megacomplexes.decay.decay_megacomplex.DecayMegacomplex` |
        |                          `foo`                          |                 `glotaran.model.megacomplex.Megacomplex`                  |
        | `glotaran.builtin.megacomplexes.decay.DecayMegacomplex` | `glotaran.builtin.megacomplexes.decay.decay_megacomplex.DecayMegacomplex` |
        """  # noqa: E501
    )

    assert f"{megacomplex_plugin_table(plugin_names=True,full_names=True)}\n" == expected



"""Test deprecations for ``glotaran.project.project``."""

from __future__ import annotations

from pathlib import Path

import pytest

from glotaran.deprecation.modules.test import deprecation_warning_on_call_test_helper
from glotaran.project.project import Project


def test_project_generate_model(tmp_path: Path):
    """Trow deprecation warning on ``Project.generate_model`` call."""
    project = Project.open(tmp_path / "deprecated_project")
    records, _ = deprecation_warning_on_call_test_helper(
        project.generate_model,
        args=("generated_test_model", "decay_parallel", {"nr_compartments": 5}),
        raise_exception=True,
    )

    assert len(records) == 1


@pytest.mark.filterwarnings(
    "ignore::glotaran.deprecation.deprecation_utils.GlotaranApiDeprecationWarning"
)
def test_project_generate_parameters(tmp_path: Path):
    """Trow deprecation warning on ``Project.generate_parameters`` call."""
    project = Project.open(tmp_path / "deprecated_project")
    project.generate_model("generated_test_model", "decay_parallel", {"nr_compartments": 5})
    records, _ = deprecation_warning_on_call_test_helper(
        project.generate_parameters, args=("generated_test_model",), raise_exception=True
    )

    assert len(records) == 1



"""This module contains the dataset group."""
from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING
from typing import Literal

import xarray as xr
from attrs import define
from attrs import field

from glotaran.model.dataset_model import DatasetModel
from glotaran.model.dataset_model import get_dataset_model_model_dimension
from glotaran.model.dataset_model import has_dataset_model_global_model
from glotaran.model.item import ModelItem
from glotaran.model.item import fill_item
from glotaran.model.item import item

if TYPE_CHECKING:
    from glotaran.model.model import Model
    from glotaran.parameter import Parameters


@item
class DatasetGroupModel(ModelItem):
    """A group of datasets which will evaluated independently."""

    residual_function: Literal[
        "variable_projection", "non_negative_least_squares"
    ] = "variable_projection"
    """The residual function to use."""

    link_clp: bool | None = None
    """Whether to link the clp parameter."""


@define
class DatasetGroup:
    """A dataset group for optimization."""

    residual_function: Literal["variable_projection", "non_negative_least_squares"]
    """The residual function to use."""

    link_clp: bool | None
    """Whether to link the clp parameter."""

    model: Model
    parameters: Parameters | None = None

    dataset_models: dict[str, DatasetModel] = field(factory=dict)

    def set_parameters(self, parameters: Parameters):
        """Set the group parameters.

        Parameters
        ----------
        parameters : Parameters
            The parameters.
        """
        self.parameters = parameters
        for label in self.dataset_models:
            self.dataset_models[label] = fill_item(
                self.model.dataset[label], self.model, parameters
            )

    def is_linkable(self, parameters: Parameters, data: Mapping[str, xr.Dataset]) -> bool:
        """Check if the group is linkable.

        Parameters
        ----------
        parameters : Parameters
            A parameter set parameters.
        data : Mapping[str, xr.Dataset]
            A the data to link.

        Returns
        -------
        bool
        """
        if any(has_dataset_model_global_model(d) for d in self.dataset_models.values()):
            return False
        dataset_models = [
            fill_item(self.model.dataset[label], self.model, parameters)
            for label in self.dataset_models
        ]
        model_dimensions = {get_dataset_model_model_dimension(d) for d in dataset_models}
        if len(model_dimensions) != 1:
            return False
        global_dimensions = set()
        for dataset in data.values():
            global_dimensions |= {
                dim for dim in dataset.data.coords if dim not in model_dimensions
            }
        return len(global_dimensions) == 1



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import xarray as xr

from glotaran.model import DatasetModel
from glotaran.model import Megacomplex
from glotaran.model import megacomplex

if TYPE_CHECKING:
    from glotaran.typing.types import ArrayLike


@megacomplex(exclusive=True)
class ClpGuideMegacomplex(Megacomplex):
    type: str = "clp-guide"
    target: str

    def calculate_matrix(
        self,
        dataset_model: DatasetModel,
        global_axis: ArrayLike,
        model_axis: ArrayLike,
        **kwargs,
    ):
        clp_label = [self.target]
        matrix = np.ones((1, 1), dtype=np.float64)
        return clp_label, matrix

    def finalize_data(
        self,
        dataset_model: DatasetModel,
        dataset: xr.Dataset,
        is_full_model: bool = False,
        as_global: bool = False,
    ):
        pass



"""Implementation of the folder Io plugin.

The current implementation is an exact copy of how ``Result.save(path)``
worked in glotaran 0.3.x and meant as an compatibility function.
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
from warnings import warn

from glotaran.deprecation import warn_deprecated
from glotaran.io import save_dataset
from glotaran.io import save_parameters
from glotaran.io import save_result
from glotaran.io.interface import ProjectIoInterface
from glotaran.plugin_system.project_io_registration import SAVING_OPTIONS_DEFAULT
from glotaran.plugin_system.project_io_registration import register_project_io

if TYPE_CHECKING:
    from glotaran.plugin_system.project_io_registration import SavingOptions
    from glotaran.project import Result


@register_project_io("legacy")
class LegacyProjectIo(ProjectIoInterface):
    """Project Io plugin to save result data in a backward compatible manner."""

    def save_result(
        self,
        result: Result,
        result_path: str,
        *,
        saving_options: SavingOptions = SAVING_OPTIONS_DEFAULT,
    ) -> list[str]:
        """Save the result to a given folder.

        Warning
        -------
        Deprecated use ``glotaran.io.save_result(result, Path(result_path) / 'result.yml')``
        instead.

        Returns a list with paths of all saved items.
        The following files are saved if not configured otherwise:
        * ``result.md``: The result with the model formatted as markdown text.
        * ``result.yml``: Yaml spec file of the result
        * ``model.yml``: Model spec file.
        * ``scheme.yml``: Scheme spec file.
        * ``initial_parameters.csv``: Initially used parameters.
        * ``optimized_parameters.csv``: The optimized parameter as csv file.
        * ``parameter_history.csv``: Parameter changes over the optimization
        * ``{dataset_label}.nc``: The result data for each dataset as NetCDF file.

        Parameters
        ----------
        result : Result
            Result instance to be saved.
        result_path : str
            The path to the folder in which to save the result.
        saving_options : SavingOptions
            Options for saving the the result.

        Returns
        -------
        list[str]
            List of file paths which were created.
        """
        warn_deprecated(
            deprecated_qual_name_usage=(
                "glotaran.io.save_result(result, result_path, format_name='legacy')"
            ),
            new_qual_name_usage=(
                "glotaran.io.save_result(result, Path(result_path) / 'result.yml')"
            ),
            to_be_removed_in_version="0.8.0",
            stacklevel=5,
        )

        return save_result(
            result=result,
            result_path=Path(result_path) / "result.yml",
            saving_options=saving_options,
            allow_overwrite=True,
        )


@register_project_io("folder")
class FolderProjectIo(ProjectIoInterface):
    """Project Io plugin to save result data to a folder.

    There won't be a serialization of the Result object, but simply
    a markdown summary output and the important data saved to files.
    """

    def save_result(
        self,
        result: Result,
        result_path: str,
        *,
        saving_options: SavingOptions = SAVING_OPTIONS_DEFAULT,
        used_inside_of_plugin: bool = False,
    ) -> list[str]:
        """Save the result to a given folder.

        Returns a list with paths of all saved items.
        The following files are saved if not configured otherwise:
        * ``result.md``: The result with the model formatted as markdown text.
        * ``initial_parameters.csv``: Initially used parameters.
        * ``optimized_parameters.csv``: The optimized parameter as csv file.
        * ``parameter_history.csv``: Parameter changes over the optimization
        * ``{dataset_label}.nc``: The result data for each dataset as NetCDF file.

        Note
        ----
        As a side effect it populates the file path properties of ``result`` which can be
        used in other plugins (e.g. the ``yml`` save_result).

        Parameters
        ----------
        result : Result
            Result instance to be saved.
        result_path : str
            The path to the folder in which to save the result.
        saving_options : SavingOptions
            Options for saving the the result.
        used_inside_of_plugin: bool
            Denote that this plugin is used from inside another plugin,
            if false a user warning will be thrown. , by default False

        Returns
        -------
        list[str]
            List of file paths which were created.

        Raises
        ------
        ValueError
            If ``result_path`` is a file.
        """
        if used_inside_of_plugin is not True:
            warn(
                UserWarning(
                    "The folder plugin is only intended for internal use by other plugins "
                    "as quick way to save most of the files. The saved result will be incomplete, "
                    "thus it is not recommended to be used directly."
                ),
                stacklevel=4,
            )

        result_folder = Path(result_path)
        if result_folder.is_file():
            raise ValueError(f"The path '{result_folder}' is not a directory.")
        result_folder.mkdir(parents=True, exist_ok=True)

        paths = []
        if saving_options.report:
            report_path = result_folder / "result.md"
            report_path.write_text(str(result.markdown()))
            paths.append(report_path.as_posix())

        initial_parameters_path = f"initial_parameters.{saving_options.parameter_format}"
        save_parameters(
            result.scheme.parameters,
            result_folder / initial_parameters_path,
            format_name=saving_options.parameter_format,
            allow_overwrite=True,
        )
        paths.append((result_folder / initial_parameters_path).as_posix())

        optimized_parameters_path = f"optimized_parameters.{saving_options.parameter_format}"
        save_parameters(
            result.optimized_parameters,
            result_folder / optimized_parameters_path,
            format_name=saving_options.parameter_format,
            allow_overwrite=True,
        )
        paths.append((result_folder / optimized_parameters_path).as_posix())

        parameter_history_path = result_folder / "parameter_history.csv"
        result.parameter_history.to_csv(parameter_history_path)
        paths.append(parameter_history_path.as_posix())

        optimization_history_path = result_folder / "optimization_history.csv"
        result.optimization_history.to_csv(optimization_history_path)
        paths.append(optimization_history_path.as_posix())

        for label, dataset in result.data.items():
            data_path = result_folder / f"{label}.{saving_options.data_format}"
            if saving_options.data_filter is not None:
                dataset = dataset[saving_options.data_filter]
            save_dataset(
                dataset,
                data_path,
                format_name=saving_options.data_format,
                allow_overwrite=True,
            )
            paths.append(data_path.as_posix())

        return paths



"""Deprecation helpers and place to put deprecated implementations till removing."""
from glotaran.deprecation.deprecation_utils import GlotaranApiDeprecationWarning
from glotaran.deprecation.deprecation_utils import GlotaranDeprecatedApiError
from glotaran.deprecation.deprecation_utils import deprecate
from glotaran.deprecation.deprecation_utils import deprecate_dict_entry
from glotaran.deprecation.deprecation_utils import deprecate_module_attribute
from glotaran.deprecation.deprecation_utils import deprecate_submodule
from glotaran.deprecation.deprecation_utils import raise_deprecation_error
from glotaran.deprecation.deprecation_utils import warn_deprecated

__all__ = [
    "deprecate",
    "deprecate_dict_entry",
    "deprecate_module_attribute",
    "deprecate_submodule",
    "raise_deprecation_error",
    "warn_deprecated",
    "GlotaranApiDeprecationWarning",
    "GlotaranDeprecatedApiError",
]



from pathlib import Path

import numpy as np
import xarray as xr

from glotaran.builtin.io.ascii.wavelength_time_explicit_file import ExplicitFile

DATA_DIR = Path(__file__).parent
TEST_FILE_ASCII = DATA_DIR.joinpath("data.ascii")

TEST_BLOCK_2x2_TOP_LEFT = np.array(
    [[9.04130489e-02, -3.17128114e-02], [-8.30020159e-02, -2.86547579e-02]]
)
TEST_BLOCK_2x2_BOTTOM_RIGHT = np.array([[0.384812891, 0.342516541], [0.350679725, 0.371651471]])


def test_read_explicit_file():
    data_file = ExplicitFile(TEST_FILE_ASCII)
    test_dataarray = data_file.read(prepare=False)
    assert isinstance(test_dataarray, xr.DataArray)
    assert test_dataarray.shape == (501, 51)
    assert np.array_equal(TEST_BLOCK_2x2_TOP_LEFT, test_dataarray[:2, :2].values.T)
    assert np.array_equal(TEST_BLOCK_2x2_BOTTOM_RIGHT, test_dataarray[-2:, -2:].values.T)
    test_dataset = data_file.read()
    assert isinstance(test_dataset, xr.Dataset)
    assert 400 == min(test_dataset.coords["spectral"].values)
    assert 600 == max(test_dataset.coords["spectral"].values)
    assert 0.0 == min(test_dataset.coords["time"].values)
    assert 5.0 == max(test_dataset.coords["time"].values)
    test_dataset.sel(spectral=[620, 630, 650], method="nearest")


def test_write_explicit_file(tmp_path):
    data_file = ExplicitFile(TEST_FILE_ASCII)
    test_dataarray_read = data_file.read(prepare=False)
    test_data_file = tmp_path.joinpath("test.ascii")
    test_data_file_write = ExplicitFile(filepath=str(test_data_file), dataset=test_dataarray_read)
    test_data_file_write.write(comment="written \n in \n test.", overwrite=True)
    test_dataarray_reread = test_data_file_write.read(prepare=False)
    assert np.array_equal(test_dataarray_read.values, test_dataarray_reread.values)



"""The glotaran parameter package."""
from glotaran.parameter.parameter import Parameter
from glotaran.parameter.parameter_history import ParameterHistory
from glotaran.parameter.parameters import Parameters


def __getattr__(attribute_name: str):
    from glotaran.deprecation import deprecate_module_attribute

    if attribute_name == "ParameterGroup":
        return deprecate_module_attribute(
            deprecated_qual_name="glotaran.parameter.ParameterGroup",
            new_qual_name="glotaran.parameter.Parameters",
            to_be_removed_in_version="0.8.0",
        )

    raise AttributeError(f"module {__name__} has no attribute {attribute_name}")



from __future__ import annotations

from pathlib import Path
from typing import Literal

import pytest

from glotaran.deprecation import GlotaranApiDeprecationWarning
from glotaran.io import save_result
from glotaran.optimization.optimize import optimize
from glotaran.project.result import Result
from glotaran.testing.simulated_data.sequential_spectral_decay import SCHEME


@pytest.fixture(scope="session")
def dummy_result():
    """Dummy result for testing."""
    print(SCHEME.data["dataset_1"])
    yield optimize(SCHEME, raise_exception=True)


@pytest.mark.parametrize("format_name", ("folder", "legacy"))
def test_save_result_folder(
    tmp_path: Path,
    dummy_result: Result,
    format_name: Literal["folder", "legacy"],
):
    """Check all files exist."""

    result_dir = tmp_path / "testresult"
    assert not result_dir.exists()
    with pytest.warns(UserWarning) as record:
        save_paths = save_result(
            result_path=str(result_dir), format_name=format_name, result=dummy_result
        )

    assert len(record) == 1
    assert Path(record[0].filename) == Path(__file__)
    if format_name == "legacy":
        assert record[0].category == GlotaranApiDeprecationWarning
    else:
        assert record[0].category == UserWarning

    wanted_files = [
        "result.md",
        "initial_parameters.csv",
        "optimized_parameters.csv",
        "parameter_history.csv",
        "dataset_1.nc",
    ]
    if format_name == "legacy":
        wanted_files += ["scheme.yml", "model.yml", "result.yml"]
    for wanted in wanted_files:
        assert (result_dir / wanted).exists()
        assert (result_dir / wanted).as_posix() in save_paths


@pytest.mark.parametrize("format_name", ("folder", "legacy"))
def test_save_result_folder_error_path_is_file(
    tmp_path: Path,
    dummy_result: Result,
    format_name: Literal["folder", "legacy"],
):
    """Raise error if result_path is a file without extension and overwrite is true."""

    result_dir = tmp_path / "testresulterror"
    result_dir.touch()

    with pytest.warns(UserWarning):
        with pytest.raises(ValueError, match="The path '.+?' is not a directory."):
            save_result(
                result_path=str(result_dir),
                format_name=format_name,
                result=dummy_result,
                allow_overwrite=True,
            )



from __future__ import annotations

from typing import TYPE_CHECKING

import numba as nb
import numpy as np
import xarray as xr
from scipy.special import erf

from glotaran.builtin.megacomplexes.decay.decay_parallel_megacomplex import DecayDatasetModel
from glotaran.builtin.megacomplexes.decay.irf import IrfMultiGaussian
from glotaran.builtin.megacomplexes.decay.util import index_dependent
from glotaran.model import DatasetModel
from glotaran.model import ItemIssue
from glotaran.model import Megacomplex
from glotaran.model import Model
from glotaran.model import ParameterType
from glotaran.model import attribute
from glotaran.model import megacomplex
from glotaran.parameter import Parameters

if TYPE_CHECKING:
    from glotaran.typing.types import ArrayLike


class OscillationParameterIssue(ItemIssue):
    def __init__(self, label: str, len_labels: int, len_frequencies: int, len_rates: int):
        self._label = label
        self._len_labels = len_labels
        self._len_frequencies = len_frequencies
        self._len_rates = len_rates

    def to_string(self) -> str:
        return (
            f"Size of labels ({self.len_labels}), frequencies ({self.len_frequencies}) "
            f"and rates ({self.len_rates}) does not match for damped oscillation "
            f"megacomplex '{self.label}'."
        )


def validate_oscillation_parameter(
    labels: list[str],
    damped_oscillation: DampedOscillationMegacomplex,
    model: Model,
    parameters: Parameters | None,
) -> list[ItemIssue]:
    issues = []

    len_labels, len_frequencies, len_rates = (
        len(damped_oscillation.labels),
        len(damped_oscillation.frequencies),
        len(damped_oscillation.rates),
    )

    if len({len_labels, len_frequencies, len_rates}) > 1:
        issues.append(
            OscillationParameterIssue(
                damped_oscillation.label, len_labels, len_frequencies, len_rates
            )
        )

    return issues


@megacomplex(dataset_model_type=DecayDatasetModel)
class DampedOscillationMegacomplex(Megacomplex):
    dimension: str = "time"
    type: str = "damped-oscillation"
    labels: list[str] = attribute(validator=validate_oscillation_parameter)
    frequencies: list[ParameterType]
    rates: list[ParameterType]

    def calculate_matrix(
        self,
        dataset_model: DatasetModel,
        global_axis: ArrayLike,
        model_axis: ArrayLike,
        **kwargs,
    ):
        clp_label = [f"{label}_cos" for label in self.labels] + [
            f"{label}_sin" for label in self.labels
        ]

        delta = np.abs(model_axis[1:] - model_axis[:-1])
        delta_min = delta[np.argmin(delta)]
        # c multiply by 0.03 to convert wavenumber (cm-1) to frequency (THz)
        # where 0.03 is the product of speed of light 3*10**10 cm/s and time-unit ps (10^-12)
        frequency_max = 1 / (2 * 0.03 * delta_min)
        frequencies = np.array(self.frequencies) * 0.03 * 2 * np.pi
        frequencies[frequencies >= frequency_max] = np.mod(
            frequencies[frequencies >= frequency_max], frequency_max
        )
        rates = np.array(self.rates)

        irf = dataset_model.irf
        matrix_shape = (
            (global_axis.size, model_axis.size, len(clp_label))
            if index_dependent(dataset_model)
            else (model_axis.size, len(clp_label))
        )
        matrix = np.ones(matrix_shape, dtype=np.float64)

        if irf is None:
            calculate_damped_oscillation_matrix_no_irf(matrix, frequencies, rates, model_axis)
        elif isinstance(irf, IrfMultiGaussian):
            if index_dependent(dataset_model):
                for i in range(global_axis.size):
                    calculate_damped_oscillation_matrix_gaussian_irf_on_index(
                        matrix[i], frequencies, rates, irf, i, global_axis, model_axis
                    )
            else:
                calculate_damped_oscillation_matrix_gaussian_irf_on_index(
                    matrix, frequencies, rates, irf, None, global_axis, model_axis
                )

        return clp_label, matrix

    def finalize_data(
        self,
        dataset_model: DatasetModel,
        dataset: xr.Dataset,
        is_full_model: bool = False,
        as_global: bool = False,
    ):
        if is_full_model:
            return

        megacomplexes = (
            dataset_model.global_megacomplex if is_full_model else dataset_model.megacomplex
        )
        unique = len([m for m in megacomplexes if isinstance(m, DampedOscillationMegacomplex)]) < 2

        prefix = "damped_oscillation" if unique else f"{self.label}_damped_oscillation"

        dataset.coords[f"{prefix}"] = self.labels
        dataset.coords[f"{prefix}_frequency"] = (prefix, self.frequencies)
        dataset.coords[f"{prefix}_rate"] = (prefix, self.rates)

        model_dimension = dataset.attrs["model_dimension"]
        global_dimension = dataset.attrs["global_dimension"]
        dim1 = dataset.coords[global_dimension].size
        dim2 = len(self.labels)
        doas = np.zeros((dim1, dim2), dtype=np.float64)
        phase = np.zeros((dim1, dim2), dtype=np.float64)
        for i, label in enumerate(self.labels):
            sin = dataset.clp.sel(clp_label=f"{label}_sin")
            cos = dataset.clp.sel(clp_label=f"{label}_cos")
            doas[:, i] = np.sqrt(sin * sin + cos * cos)
            phase[:, i] = np.unwrap(np.arctan2(sin, cos))

        dataset[f"{prefix}_associated_spectra"] = (
            (global_dimension, prefix),
            doas,
        )

        dataset[f"{prefix}_phase"] = (
            (global_dimension, prefix),
            phase,
        )

        if index_dependent(dataset_model):
            dataset[f"{prefix}_sin"] = (
                (
                    global_dimension,
                    model_dimension,
                    prefix,
                ),
                dataset.matrix.sel(clp_label=[f"{label}_sin" for label in self.labels]).values,
            )

            dataset[f"{prefix}_cos"] = (
                (
                    global_dimension,
                    model_dimension,
                    prefix,
                ),
                dataset.matrix.sel(clp_label=[f"{label}_cos" for label in self.labels]).values,
            )
        else:
            dataset[f"{prefix}_sin"] = (
                (model_dimension, prefix),
                dataset.matrix.sel(clp_label=[f"{label}_sin" for label in self.labels]).values,
            )

            dataset[f"{prefix}_cos"] = (
                (model_dimension, prefix),
                dataset.matrix.sel(clp_label=[f"{label}_cos" for label in self.labels]).values,
            )


@nb.jit(nopython=True, parallel=True)
def calculate_damped_oscillation_matrix_no_irf(matrix, frequencies, rates, axis):
    idx = 0
    for frequency, rate in zip(frequencies, rates):
        osc = np.exp(-rate * axis - 1j * frequency * axis)
        matrix[:, idx] = osc.real
        matrix[:, idx + 1] = osc.imag
        idx += 2


def calculate_damped_oscillation_matrix_gaussian_irf_on_index(
    matrix: ArrayLike,
    frequencies: ArrayLike,
    rates: ArrayLike,
    irf: IrfMultiGaussian,
    global_index: int | None,
    global_axis: ArrayLike,
    model_axis: ArrayLike,
):
    centers, widths, scales, shift, _, _ = irf.parameter(global_index, global_axis)
    for center, width, scale in zip(centers, widths, scales):
        matrix += calculate_damped_oscillation_matrix_gaussian_irf(
            frequencies,
            rates,
            model_axis,
            center,
            width,
            shift,
            scale,
        )
    matrix /= np.sum(scales)


def calculate_damped_oscillation_matrix_gaussian_irf(
    frequencies: np.ndarray,
    rates: np.ndarray,
    model_axis: np.ndarray,
    center: float,
    width: float,
    shift: float,
    scale: float,
):
    """Calculate the damped oscillation matrix taking into account a gaussian irf

    Parameters
    ----------
    frequencies : np.ndarray
        an array of frequencies in THz, one per oscillation
    rates : np.ndarray
        an array of rates, one per oscillation
    model_axis : np.ndarray
        the model axis (time)
    center : float
        the center of the gaussian IRF
    width : float
        the width (σ) parameter of the the IRF
    shift : float
        a shift parameter per item on the global axis
    scale : float
        the scale parameter to scale the matrix by

    Returns
    -------
    np.ndarray
        An array of the real and imaginary part of the oscillation matrix,
        the shape being (len(model_axis), 2*len(frequencies)), with the first
        half of the second dimension representing the real part,
        and the other the imagine part of the oscillation
    """
    shifted_axis = model_axis - center - shift
    # For calculations using the negative rates we use the time axis
    # from the beginning up to 5 σ from the irf center
    left_shifted_axis_indices = np.where(shifted_axis < 5 * width)[0]
    left_shifted_axis = shifted_axis[left_shifted_axis_indices]
    neg_idx = np.where(rates < 0)[0]
    # For calculations using the positive rates axis we use the time axis
    # from 5 σ before the irf center until the end
    right_shifted_axis_indices = np.where(shifted_axis > -5 * width)[0]
    right_shifted_axis = shifted_axis[right_shifted_axis_indices]
    pos_idx = np.where(rates >= 0)[0]

    d = width**2
    k = rates + 1j * frequencies
    dk = k * d
    sqwidth = np.sqrt(2) * width

    a = np.zeros((len(model_axis), len(rates)), dtype=np.complex128)
    a[np.ix_(right_shifted_axis_indices, pos_idx)] = np.exp(
        (-1 * right_shifted_axis[:, None] + 0.5 * dk[pos_idx]) * k[pos_idx]
    )

    a[np.ix_(left_shifted_axis_indices, neg_idx)] = np.exp(
        (-1 * left_shifted_axis[:, None] + 0.5 * dk[neg_idx]) * k[neg_idx]
    )

    b = np.zeros((len(model_axis), len(rates)), dtype=np.complex128)
    b[np.ix_(right_shifted_axis_indices, pos_idx)] = 1 + erf(
        (right_shifted_axis[:, None] - dk[pos_idx]) / sqwidth
    )
    # For negative rates we flip the sign of the `erf` by using `-sqwidth` in lieu of `sqwidth`
    b[np.ix_(left_shifted_axis_indices, neg_idx)] = 1 + erf(
        (left_shifted_axis[:, None] - dk[neg_idx]) / -sqwidth
    )

    osc = a * b * scale

    return np.concatenate((osc.real, osc.imag), axis=1)



import xarray as xr

from glotaran.testing.simulated_data.parallel_spectral_decay import DATASET as parallel_dataset
from glotaran.testing.simulated_data.sequential_spectral_decay import DATASET as sequential_dataset


def test_dataset():
    assert isinstance(parallel_dataset, xr.Dataset)
    assert isinstance(sequential_dataset, xr.Dataset)



"""Test deprecations for ``glotaran.project.project``."""

from __future__ import annotations

from pathlib import Path

import pytest

from glotaran.deprecation import GlotaranApiDeprecationWarning
from glotaran.optimization.optimize import optimize
from glotaran.testing.simulated_data.sequential_spectral_decay import SCHEME


def test_result_number_of_data_points():
    """Trow deprecation warning on accessing ``Result.number_of_data_points``."""
    print(SCHEME.data["dataset_1"])
    result = optimize(SCHEME, raise_exception=True)
    with pytest.warns(GlotaranApiDeprecationWarning) as records:
        result.number_of_data_points

        assert len(records) == 1
        assert Path(records[0].filename) == Path(
            __file__
        ), f"{Path(records[0].filename)=}, {Path(__file__)=}"


def test_result_number_of_parameters():
    """Trow deprecation warning on accessing ``Result.number_of_parameters``."""
    print(SCHEME.data["dataset_1"])
    result = optimize(SCHEME, raise_exception=True)
    with pytest.warns(GlotaranApiDeprecationWarning) as records:
        result.number_of_parameters

        assert len(records) == 1
        assert Path(records[0].filename) == Path(
            __file__
        ), f"{Path(records[0].filename)=}, {Path(__file__)=}"
